#========================================================================================
# §package      DE_EPAGES::FacebookBusiness::API::Object::FacebookServerClient
# §state        public
#----------------------------------------------------------------------------------------
# §description  object for calling functions on the FacebookServer
#========================================================================================
package DE_EPAGES::FacebookBusiness::API::Object::FacebookServerClient;

use strict;
use Digest::SHA qw( hmac_sha256_base64 );
use Encode qw ( encode );
use HTTP::Request;
use JSON::XS qw ( encode_json );
use JSON qw( to_json from_json);
use LWP::UserAgent;
use MIME::Base64;
use DE_EPAGES::Core::API::Error          qw( Error ExistsError ResetError );
use DE_EPAGES::Core::API::Log            qw( GetCommunicationLog );
use DE_EPAGES::Core::API::TestParameter  qw( TestString TestObject );
use DE_EPAGES::FacebookBusiness::API::Constants qw( FACEBOOK_SERVER_URL FACEBOOK_SERVER_REQUEST_SECRET );
use DE_EPAGES::Object::API::Factory      qw( InsertObject );

#ERRORDEFINITION
sub ErrorDefinition { return {}
}
#ERRORDEFINITION END

#========================================================================================
sub new {
    my $class = shift;
    my ($Shop)  = @_;
    my $self  = bless {
        'Shop' => $Shop
    }, $class;

    TestObject('Shop', $self->{'Shop'}, 'DE_EPAGES::Shop::API::Object::Shop');

    return $self;
}

#========================================================================================
# §function     call
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->call();
#----------------------------------------------------------------------------------------
# §description  prepares a signed request for an api call.
#----------------------------------------------------------------------------------------
# §input        $CallName | signed data for sending to the epages facebook server | string
# §input        $aParams | api call name | array
# §return       $hResponse |  response from epages fb server | hash
#========================================================================================
sub call {
    my $self = shift;
    my ($CallName, $aParams) = @_;

    # use constants!!!
    my $hParams = {};

    $hParams->{'call'} = $CallName;
    $hParams->{'params'} = $aParams;

    #DE_EPAGES::Core::API::Log::LogDebug('FBSClient Call: ' . $hParams->{'call'}, $hParams->{'params'}, 'DEBUG');

    my $JSONParams = to_json($hParams,'utf_8');
    my $Base64JSONParams = encode_base64($JSONParams);
    my $Signature = hmac_sha256_base64($Base64JSONParams, FACEBOOK_SERVER_REQUEST_SECRET);
    my $SignedData = $Signature . '.' . $Base64JSONParams;

    my $Response = $self->sendRequest($SignedData);

    #DE_EPAGES::Core::API::Log::LogDebug('FBSClient Call Response: ' . $hParams->{'call'}, $Response, 'DEBUG');

    return $Response;

}

#========================================================================================
# §function     sendRequest
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->sendRequest();
#----------------------------------------------------------------------------------------
# §description  Sends a request to the epages facebook server.
#----------------------------------------------------------------------------------------
# §input        $SignedData | signed data for sending to the epages facebook server | string
# §return       $hResponse |  response from epages fb server | hash
#========================================================================================
sub sendRequest {
    my $self = shift;
    my ($SignedData) = @_;

    TestString( 'SignedData', $SignedData );

    my $Method = 'POST';
    my $Request = HTTP::Request->new($Method => FACEBOOK_SERVER_URL);

    my $hParams = {};
    $hParams->{'Header'} = {};
    $hParams->{'Content'}  = {
                                'signed_data' => $SignedData,
                             };

    my $hHTTPHeader = {
                        'Content-Type'  => 'application/json',
                        'Accept'        => 'application/json',
                        'Accept-Charset'=> 'utf-8',
                     };

    # add additional header values
    foreach my $Key (keys(%{  $hParams->{'Header'} })) {
        $hHTTPHeader->{$Key} = $hParams->{'Header'}->{$Key};
    }
    $Request->header(%$hHTTPHeader);

    my $EncodedContent;
    $EncodedContent = encode('UTF-8',encode_json($hParams->{'Content'})) if defined $hParams->{'Content'};

    # adds the content and send request
    $Request->content($EncodedContent) if defined $hParams->{'Content'};

    my $LogCategory = $self->_getLogCategory();
    $self->logCommunication($LogCategory, $Request);
    my $Response = LWP::UserAgent->new->request($Request);
    $self->logCommunication( $LogCategory, $Response);

    # error handling if the FB server call fails
    unless ($Response->is_success) {
        $self->_doError($Response, $hParams, $Method);
    }

    # return decoded content if we have a GET request only
    my $JSONResponse = $Response->decoded_content;
    my $hResponse = $JSONResponse ? from_json($JSONResponse, {utf8 => 1}) : {};

    # test if we have an error from the direct FB call
    if((ref $hResponse eq 'HASH') and $hResponse->{'error'}) {
        $self->_doError($hResponse, $hParams, $Method);
    }

    return $hResponse;

}

#========================================================================================
# §function     _doError
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_doError();
#----------------------------------------------------------------------------------------
# §description  Throws an API_CALL_ERROR error.
#----------------------------------------------------------------------------------------
# §input        $Response | response | string
# §input        $hParams | parameters | string
# §input        $Method | method | string
#========================================================================================
sub _doError {
    my $self = shift;
    my ($Response, $hParams, $Method) = @_;

    ResetError() if ExistsError();
    my $ErrorObject = {
        'Message'  =>  "API call to FB failed for method: $Method",
        'Method'   =>  $Method,
        'ApiURL'   =>  FACEBOOK_SERVER_URL,
        'Content'  =>  $hParams->{'Content'},
        'Response' =>  $Response,
    };
    Error('API_CALL_ERROR', $ErrorObject);

    return;

}

#========================================================================================
# §function     _getLogCategory
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_getLogCategory();
#----------------------------------------------------------------------------------------
# §description  Returns the log category if logging is enabled.
#----------------------------------------------------------------------------------------
# §return       $LogCategory |  log category for enabled logging | string
#========================================================================================
sub _getLogCategory {
    my $self = shift;

    return  "FacebookBusiness::FacebookServerClient";
}

#========================================================================================
# §function     logCommunication
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $RemoteFacebook->logCommunication( $Category, $HTTPObject );
# §example      $RemoteFacebook->logCommunication( 'FacebookBusiness::API::Object::FacebookServerClient', $Response );
#----------------------------------------------------------------------------------------
# §description  Logs HTTP communications.
#----------------------------------------------------------------------------------------
# §input        $Category | log category, undef for no logging | string
# §input        $Message | Message to be logged | HTTP::Request or HTTP::Response object
#========================================================================================
sub logCommunication {
    shift;
    my ($Category, $Message) = @_;

    my $Logger = GetCommunicationLog($Category);

    if (defined $Category && $Logger->is_debug) {
            if( $Message->isa('HTTP::Request') ) {
                $Logger->debug( "Request headers:\n" . $Message->headers_as_string );
                $Logger->debug( $Message->method .' '. $Message->uri ."\n". $Message->content);
            }
            else {
                $Logger->debug( "Response headers:\n" . $Message->headers_as_string );
            }
    }
    return;
}

#========================================================================================
# §function     shopEnvironment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Environment = $self->shopEnvironment();
#----------------------------------------------------------------------------------------
# §description  Returns the Facebook Shop Environment by the given shop.
#               Facebook Shop Environment is created if not exists
#----------------------------------------------------------------------------------------
# §return       $Environment | Facebook Shop Environment | Object
#========================================================================================
sub shopEnvironment {
    my $self = shift;
    my $Shop = $self->{'Shop'};

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    return $Shop->child('FacebookBusinessShopEnvironment') if $Shop->existsChild('FacebookBusinessShopEnvironment');
    return InsertObject('FacebookBusinessShopEnvironment', {
                                                            'Parent' => $Shop,
                                                            'Alias'  => 'FacebookBusinessShopEnvironment',
                                                           });
}

#========================================================================================
# §function     existsShopEnvironment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $EnvironmentExists = $self->existsShopEnvironment();
#----------------------------------------------------------------------------------------
# §description  Test if Facebook Shop Environment exists.
#----------------------------------------------------------------------------------------
# §return       $EnvironmentExists | Facebook Shop Environment exists or not | boolean
#========================================================================================
sub existsShopEnvironment {
    my $self = shift;
    my $Shop = $self->{'Shop'};

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    return $Shop->existsChild('FacebookBusinessShopEnvironment');

}

#========================================================================================
# §function     createStateInfo
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $hStateInfo = $self->createStateInfo($hBMInfo);
#----------------------------------------------------------------------------------------
# §description  Returns states about the installed and added business elements in the shop
#               dependent from the existing business elements on Facebook
#----------------------------------------------------------------------------------------
# §input        $hBMInfo | business manager information | ref.hash
# §return       $hStateInfo | state information | ref.hash
#========================================================================================
sub createStateInfo {
    my $self = shift;
    my ($hBMInfo) = @_;

    my $hStateInfo = {};

   #add BM
    my $BMCount = 0;
    my $AACount = 0;
    my $PXCount = 0;
    foreach my $hBM (@{$hBMInfo->{'data'}}) {
        my $hElement = {};
        $BMCount ++;

        foreach my $key (qw(id name link)){
            $hElement->{$key} = $hBM->{$key};
        }
        push @{$hStateInfo->{'BM'}}, $hElement;

        #add Ad accounts
        foreach my $hBM (@{$hBM->{'owned_ad_accounts'}->{'data'}}) {
            my $hAAElement = {};
            $AACount ++;
            foreach my $key (qw(id account_id)){
                $hAAElement->{$key} = $hBM->{$key};
            }
            push @{$hStateInfo->{'AA'}}, $hAAElement;
        }
        #add Ad Pixels
        foreach my $hBM (@{$hBM->{'adspixels'}->{'data'}}) {
            my $hPElement = {};
            $PXCount ++;
            foreach my $key (qw(id)){              
                $hPElement->{$key} = $hBM->{$key};
            }
            push @{$hStateInfo->{'PX'}}, $hPElement;
        }

    }
    $hStateInfo->{'BMcount'} = $BMCount;
    $hStateInfo->{'AAcount'} = $AACount;
    $hStateInfo->{'PXcount'} = $PXCount;

    return $hStateInfo;
}

1;
