#========================================================================================
# package      DE_EPAGES::MolliePayments::t::unit::Requests_Orders
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::MolliePayments::API::Requests::Orders
#========================================================================================
package DE_EPAGES::MolliePayments::t::unit::Requests_Orders;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::MolliePayments::API::Tools qw ( FormatFloatForMollie );

use DE_EPAGES::Test::Mock::MockifyEpagesObject;

use Test::Mockify::Sut;
use Test::More;

use strict;

use constant ACCESS_TOKEN => 'MyAccessToken';

use constant ADDRESS_CITY => 'Bcken';

use constant ADDRESS_COUNTRY => {
    'Code2' => 'DE',
    'Code3' => 'DEU',
    'Native' => 'Deutschland'
};

use constant ADDRESS_EMAIL => 'mario.schafer@example.com';

use constant ADDRESS_FIRST_NAME => 'Mario';

use constant ADDRESS_LAST_NAME => 'Schfer';

use constant ADDRESS_ZIP_CODE => '27333';

use constant ADDRESS_STREET => 'Karl-Liebknecht-Strasse 75';

use constant BASKET_CURRENCY => 'EUR';

use constant INITIAL_ORDER_NUMBER_NOT_SET => '-';

use constant LANGUAGE_ID_DE => 1;

use constant BASKET_LOCALE_NOT_IN_MOLLLIE => 'de_LU'; # Deutsch language from Luxemburg

use constant BASKET_LOCALE_IN_MOLLIE => 'de_DE';

use constant BASKET_TOTAL_AMOUNT => 18.00;

use constant MOLLIE_PAYMENT_METHOD => 'MyMolliePaymentMethod';

use constant ORDERS_SUFFIX_EPAGESJ_PATH => '/orders';

use constant ORDER_REDIRECT_URI => 'https://myshop.epages.de/epages/myshop.sf/';

use constant PRODUCT_ALIAS => 'LineItemProduct_GUID';

use constant PRODUCT_BASE_PRICE => 10.25;
use constant PRODUCT_BASE_PRICE_WHEN_DECIMAL_QUANTITY => 20.50;

use constant PRODUCT_QUANTITY_INTEGER => 2;
use constant PRODUCT_QUANTITY_DECIMAL => 2.14;

use constant PRODUCT_DISCOUNT => 2.50;

use constant PRODUCT_NAME => 'MyProductName';

use constant PRODUCT_TAX_RATE => 0.19;

use constant PRODUCT_TOTAL_PRICE => 18.00;

use constant PROFILE_ID => 'MyProfileId';

use constant RESPONSE_HASH => {
    'id' => 'MollieOrderId',
    'mode' => 'test',
    'status' => 'Created',
    'amount' => {
        'currency' => 'EUR',
        'value' => '26.00'
    },
    'redirectUrl' => ORDER_REDIRECT_URI,
    'webhookUrl' => 'https://example.org/webhook',
    'lines' => [],
    '_links' => {
        'self' => {
            'href' => 'https://api.mollie.com/v2/orders/order_id',
            'type' => 'application/hal+json'
        },
        'checkout' => {
            'href' => 'https://www.mollie.com/payscreen/order/checkout/order_id',
            'type' => 'text/html'
        },
        'documentation' => {
            'href' => 'https://docs.mollie.com/reference/v2/orders-api/get-order',
            'type' => 'text/html'
        }
    }
};

use constant RESPONSE_HASH_FILTERED => {
    'id' => 'MollieOrderId',
    'mode' => 'test',
    'status' => 'Created',
    'amount' => {
        'currency' => 'EUR',
        'value' => '26.00'
    },
    'checkout' => {
        'href' => 'https://www.mollie.com/payscreen/order/checkout/order_id',
        'type' => 'text/html'
    }
};

use constant SHOP_NAME => 'MyShopName';

my $ProductQuantity = PRODUCT_QUANTITY_INTEGER;

sub test {
    my $self = shift;

    $self->test_Order_creation();
    $self->test_Order_provision_for_post_path_and_headers();
    $self->test_Order_provision_for_post_payload();
    $self->test_Order_provision_for_post_payload_with_decimal_product_quantity();
    $self->test_Order_provision_for_shop();
    $self->test_Order_handling_for_response();

    done_testing();
}

#----------------------------------------------------------------------------------------
# Tests
#----------------------------------------------------------------------------------------
sub test_Order_creation {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $SUT = $self->_createSUT();
    is( scalar(keys %{$SUT->{'data'}}), 3, "$SubTestName - Orders object gets initialized with some arguments.");

    is( scalar(keys %{$SUT->{'payload'}}), 0, "$SubTestName - Orders object gets initialized with an empty payload.");
}

sub test_Order_provision_for_post_path_and_headers {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $SUT = $self->_createSUT();
    is(
            $SUT->getEpagesjRequestSuffixUrlPath('POST'),
            ORDERS_SUFFIX_EPAGESJ_PATH,
            "$SubTestName - Orders object provides epagesj suffix path."
    );

    is_deeply(
            $SUT->getHttpHeaders('POST'),
            { 'X-AccessToken' => ACCESS_TOKEN },
            "$SubTestName - Orders object provides HTTP headers."
    );
}

sub test_Order_provision_for_post_payload {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $SUT = $self->_createSUT();
    my $hPayload = $SUT->getPayload('POST');

    is( $hPayload->{'profileId'}, PROFILE_ID, "$SubTestName - Orders payload profile id is retrieved.");

    is( $hPayload->{'method'}, MOLLIE_PAYMENT_METHOD, "$SubTestName - Orders payload payment method is retrieved.");

    is_deeply(
            $hPayload->{'amount'},
            { 'value' => FormatFloatForMollie(BASKET_TOTAL_AMOUNT), 'currency' => BASKET_CURRENCY },
            "$SubTestName - Order basket amount is retrieved."
    );

    is( $hPayload->{'locale'}, BASKET_LOCALE_IN_MOLLIE, "$SubTestName - Orders payload available locale code is retrieved.");

    is( $hPayload->{'orderNumber'}, INITIAL_ORDER_NUMBER_NOT_SET, "$SubTestName - Orders payload order number retrieved.");

    is( $hPayload->{'redirectUrl'}, ORDER_REDIRECT_URI, "$SubTestName - Orders payload redirect URI retrieved.");

    is_deeply(
            $hPayload->{'lines'},
            [{
                'name' => PRODUCT_NAME,
                'quantity' => PRODUCT_QUANTITY_INTEGER,
                'unitPrice' => {
                    'value' => FormatFloatForMollie(PRODUCT_BASE_PRICE),
                    'currency' => BASKET_CURRENCY
                },
                'totalAmount' => {
                    'value' => FormatFloatForMollie(PRODUCT_TOTAL_PRICE),
                    'currency' => BASKET_CURRENCY
                },
                'discountAmount' => {
                    'value'    => FormatFloatForMollie(PRODUCT_DISCOUNT),
                    'currency' => BASKET_CURRENCY
                },
                'vatRate' => '19.00',
                'vatAmount' => {
                    'value' => '2.87', # totalAmount  (vatRate / (100 + vatRate))
                    'currency' => BASKET_CURRENCY
                },
                'metadata' => {
                    'alias' => PRODUCT_ALIAS
                }
            }],
            "$SubTestName - Orders payload line items retrieved.");
}

sub test_Order_provision_for_post_payload_with_decimal_product_quantity {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    $ProductQuantity = PRODUCT_QUANTITY_DECIMAL;
    my $SUT = $self->_createSUT();
    my $hPayload = $SUT->getPayload('POST');

    is( $hPayload->{'profileId'}, PROFILE_ID, "$SubTestName - Orders payload profile id is retrieved.");

    is( $hPayload->{'method'}, MOLLIE_PAYMENT_METHOD, "$SubTestName - Orders payload payment method is retrieved.");

    is_deeply(
            $hPayload->{'amount'},
            { 'value' => FormatFloatForMollie(BASKET_TOTAL_AMOUNT), 'currency' => BASKET_CURRENCY },
            "$SubTestName - Order basket amount is retrieved."
    );

    is( $hPayload->{'locale'}, BASKET_LOCALE_IN_MOLLIE, "$SubTestName - Orders payload available locale code is retrieved.");

    is( $hPayload->{'orderNumber'}, INITIAL_ORDER_NUMBER_NOT_SET, "$SubTestName - Orders payload order number retrieved.");

    is( $hPayload->{'redirectUrl'}, ORDER_REDIRECT_URI, "$SubTestName - Orders payload redirect URI retrieved.");

    is_deeply(
            $hPayload->{'lines'},
            [{
                'name' => PRODUCT_NAME,
                'quantity' => 1, # Quantity set to 1 on decimal quantities
                'unitPrice' => {
                    'value' => FormatFloatForMollie(20.50), # unit price changes to accommodate to quantity value changes
                    'currency' => BASKET_CURRENCY
                },
                'totalAmount' => {
                    'value' => FormatFloatForMollie(PRODUCT_TOTAL_PRICE),
                    'currency' => BASKET_CURRENCY
                },
                'discountAmount' => {
                    'value'    => FormatFloatForMollie(PRODUCT_DISCOUNT),
                    'currency' => BASKET_CURRENCY
                },
                'vatRate' => '19.00',
                'vatAmount' => {
                    'value' => '2.87', # totalAmount  (vatRate / (100 + vatRate))
                    'currency' => BASKET_CURRENCY
                },
                'metadata' => {
                    'alias' => PRODUCT_ALIAS
                }
            }],
            "$SubTestName - Orders payload line items retrieved.");
}

sub test_Order_provision_for_shop {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $SUT = $self->_createSUT();
    my $Shop = $SUT->getShop();

    is( $Shop->alias(), SHOP_NAME, "$SubTestName - Orders object provides shop object.");
}

sub test_Order_handling_for_response {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $SUT = $self->_createSUT();

    is_deeply(
            $SUT->handleResponse(RESPONSE_HASH),
            RESPONSE_HASH_FILTERED,
            "$SubTestName - Orders response filtering provides expected response."
    );
}


#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------
sub _createSUT {
    my $self = shift;

    my $hData = $self->_generateRequestData();
    my $aParameterList = [$hData];
    my $MockifySut = Test::Mockify::Sut->new('DE_EPAGES::MolliePayments::API::Requests::Orders', $aParameterList);

    return $MockifySut->getMockObject();
}

sub _generateRequestData {
    my $self = shift;

    return {
        'basket' => $self->_createBasket(),
        'payment' => $self->_createPaymentMethod(),
        'redirectUrl' => ORDER_REDIRECT_URI
    };
}

sub _createBasket {
    my $self = shift;

    my $hAttributes = {
        'BillingAddress' => $self->_createAddress(),
        'LineItemContainer' => $self->_createLineItemContainer(),
        'ShippingAddress' => undef
    };
    my $MockifyBasket = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Order::API::Object::Basket', $hAttributes
    );

    return $MockifyBasket->getMockObject();
}

sub _createAddress {
    my $self = shift;

    my $hAttributes = {
        'City' => ADDRESS_CITY,
        'Country' => ADDRESS_COUNTRY,
        'EMail' => ADDRESS_EMAIL,
        'FirstName' => ADDRESS_FIRST_NAME,
        'LastName' => ADDRESS_LAST_NAME,
        'State' => undef,
        'Street' => ADDRESS_STREET,
        'Zipcode' => ADDRESS_ZIP_CODE
    };
    my $MockifyBasket = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Address::API::Object::Address', $hAttributes
    );

    return $MockifyBasket->getMockObject();
}

sub _createLineItemContainer {
    my $self = shift;

    my $hAttributes = {
        'CurrencyID' => BASKET_CURRENCY,
        'GrandTotal' => BASKET_TOTAL_AMOUNT,
        'LanguageID' => LANGUAGE_ID_DE,
        'LocaleID' => BASKET_LOCALE_NOT_IN_MOLLLIE,
    };
    my $MockifyContainer = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Order::API::Object::LineItemContainer', $hAttributes
    );
    $MockifyContainer->addMockWithReturnValue('children', [ $self->_createLineItemProduct() ]);

    return $MockifyContainer->getMockObject();
}

sub _createLineItemProduct {
    my $self = shift;

    my $hAttributes = {
        'Alias' => PRODUCT_ALIAS,
        'BasePrice' => PRODUCT_BASE_PRICE,
        'LineItemPrice' => PRODUCT_TOTAL_PRICE,
        'Discount' => PRODUCT_DISCOUNT,
        'NameOrAlias' => PRODUCT_NAME,
        'Quantity' => $ProductQuantity,
        'TaxRate' => PRODUCT_TAX_RATE,
    };
    my $MockifyLineItem = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Product::API::Object::LineItemProduct', $hAttributes
    );
    $MockifyLineItem->addMock('instanceOf', sub {
            shift;
            my ($ClassName) = @_;
            return 1 if ($ClassName eq 'LineItemProduct');
            return 0;
    });

    return $MockifyLineItem->getMockObject();
}

sub _createPaymentMethod {
    my $self = shift;

    my $hAttributes = {
        'AccessToken' => ACCESS_TOKEN,
        'IsInTestMode' => 1,
        'Method' => MOLLIE_PAYMENT_METHOD,
        'ProfileId' => PROFILE_ID,
        'Site' => $self->_createShop()
    };
    my $MockifyPaymentMethod = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::MolliePayments::API::Object::PaymentMethodMolliePayments', $hAttributes
    );

    return $MockifyPaymentMethod->getMockObject();
}

sub _createShop {
    my $self = shift;

    my $hAttributes = {
        'Alias' => SHOP_NAME
    };
    my $MockifyPaymentMethod = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Shop::API::Object::Shop', $hAttributes
    );

    return $MockifyPaymentMethod->getMockObject();
}

#========================================================================================
# Main
#========================================================================================

__PACKAGE__->runScript();
