#==============================================================================
# package      DE_EPAGES::Moneybookers::UI::Shop
# state        private
#==============================================================================
package DE_EPAGES::Moneybookers::UI::Shop;
use base DE_EPAGES::Order::UI::Shop;

use strict;
use DE_EPAGES::Object::API::Factory       qw (ExistsObject ExistsObjectByGUID LoadObjectByGUID);
use DE_EPAGES::Core::API::Error           qw (Error ExistsError GetError);
use DE_EPAGES::Core::API::Warning         qw (Warning);
use DE_EPAGES::Core::API::PerlTools       qw (GetCurrentDateTime);
use DE_EPAGES::ExternalPayment::API::Log  qw (LogPayment);
use DE_EPAGES::Moneybookers::API::Tools;
use DE_EPAGES::Order::DAL::Basket         qw (DAL_UnCachedInfoBasket);
use DE_EPAGES::Order::UI::Basket;
use Digest::MD5 'md5_hex';
use Encode qw(encode_utf8);
use HTTP::Status qw (RC_OK RC_CONFLICT);

sub ErrorDefinition {
    return {
        'MoneybookersNoBasket'           => 'No Basket Found for TransactionID #TransactionID. (ShopID:#ShopID)',
        'MoneybookersSignatureMismatch'  => 'Signature: #Signature does not match the calculated Signature: #Mac  (ShopID:#ShopID MBTransactionID:#MBTransactionID TransactionID:#TransactionID)',
    }
}


#========================================================================================
# function     PaymentCallbackMoneybookers
# state        private
#----------------------------------------------------------------------------------------
# syntax       PaymentCallbackMoneybookers($Servlet)
#----------------------------------------------------------------------------------------
# description  PaymentCallbackMoneybookers is called if the customer click on the
#               Return Button. If no error occurs the order already exists at this point
#               (created by PaymentStatusMoneybookers ).
#               If the order not exists, the order is created
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet |  object
#========================================================================================
sub PaymentCallbackMoneybookers {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop       = $Servlet->object();
    my $Form       = $Servlet->form();
    my $ServletParams = $Servlet->params();

    my $LineItemContainer = $self->_validateCallParametersAndGetContainer($Shop, $Form, $ServletParams);

    my $LineItemPayment = $LineItemContainer->get('Payment');
    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');

    my $ObjectPath = $PaymentMethod->pathString() . ' (ObjectID ' . $PaymentMethod->id() . ')';
    $self->_logInformationWithShopContext(
        $Shop,
        "Moneybookers callback for $ObjectPath; form values",
        $ServletParams
    );

    #-- this is Basket or CustomerOrder
    my $ParentObject = $LineItemContainer->parent();

    #-- create order and show order confirmation page
    if ($ParentObject->instanceOf('Basket') and $PaymentMethod->instanceOf('PaymentMethodMoneybookers')) {
        my $Order;
        my $OrderGUID = DAL_UnCachedInfoBasket($ParentObject->id())->{'OrderGUID'};
        if( defined $OrderGUID ) {
            if( ExistsObjectByGUID($OrderGUID) ) {
                # order already exist
                $Order = LoadObjectByGUID($OrderGUID);
            }
            else {
                # order is beeing created by another process
                # show a page with a refresh button
                $Servlet->vars('ObjectID', $ParentObject->id());
                Error('INVALID_FORM', {
                    'ViewVars'=> {
                        'FormError' => 0,
                        'IsBlocked' => 1,
                        'ObjectID', $ParentObject->id(),
                        'OrderGUID' => $OrderGUID,
                    },
                    'ViewAction' => 'ViewBlocked',
                });

            }
        }
        else {
            my $status = $LineItemPayment->get('TransStatus');
            if (defined($status) && ($status eq '0' || $status eq '2')) {
                # create order
                $Order = DE_EPAGES::Order::UI::Basket->doBasket2Order($Servlet, $ParentObject);
                $Servlet->vars('ViewObjectID', $Order->id());
                $Servlet->vars('ViewAction', $Shop->isUnity() ? 'ViewMultiCheckoutCustomerOrder' : undef);
                $Servlet->registerSession($Order->get('User'), 0);
            }
            else {
                # show declined page
                $Servlet->vars('ViewObjectID', $Shop->id());
                if ($Shop->isUnity()) {
                    $LineItemPayment->set( { 'PaymentErrorCode' => 'Failed' } );
                    $Servlet->vars('ViewAction',  'ViewMultiCheckoutCustomerOrder');
                } else {
                    $Servlet->vars('ViewAction',  'ViewMoneybookersDeclined');
                }
                $Servlet->vars('MoneybookersBasket', $ParentObject->tleHash());
                $Servlet->registerSession($ParentObject->get('User'), 0);
            }
        }
    }
    else {
        #-- show order confirmation page
        $Servlet->vars('ViewObjectID', $ParentObject->id());
        $Servlet->vars('ViewAction', $Shop->isUnity() ? 'ViewMultiCheckoutCustomerOrder' : undef);
    }
    return;
}


#========================================================================================
# function     PaymentMoneybookersCancel
# state        public
#----------------------------------------------------------------------------------------
# syntax       PaymentMoneybookersCancel($Servlet)
#----------------------------------------------------------------------------------------
# description  Is invoked once a customer canceled the payment with Skill or the
#               payment failed and the customer is redirected to the shop.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet |  object
#========================================================================================
sub PaymentMoneybookersCancel {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $ServletParams = $Servlet->params();

    my $LineItemContainer = $self->_validateCallParametersAndGetContainer($Shop, $Form, $ServletParams);

    my $LineItemPayment = $LineItemContainer->get('Payment');
    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');

    my $ObjectPath = $PaymentMethod->pathString() . ' (ObjectID ' . $PaymentMethod->id() . ')';
    $self->_logInformationWithShopContext(
        $Shop,
        "Moneybookers cancellation for $ObjectPath; form values",
        $ServletParams
    );
    #-- this is Basket or CustomerOrder
    my $BasketOrder = $LineItemContainer->parent();

    $Servlet->vars('ViewObjectID', $BasketOrder->id());
    if($Shop->isUnity()) {
        $LineItemPayment->set({'PaymentErrorCode' => 'Canceled'});
        $Servlet->vars('ViewAction', 'ViewMultiCheckoutCustomerOrder');
    } else {
        $Servlet->vars('ViewAction', 'ViewPaymentMoneybookersCancel');
    }

    return;
}

#========================================================================================
# function     PaymentStatusMoneybookers
# state        private
#----------------------------------------------------------------------------------------
# syntax       PaymentStatusMoneybookers($Servlet)
#----------------------------------------------------------------------------------------
# description  PaymentStatusMoneybookers is called if the payment was ok
#               (internal request in MB)
#               When the payment process is complete Moneybookers' payment server will post the
#               transaction's details to the 'status_url' URL provided by the merchant.
#               The Moneybookers server will continue to post the status reports until a
#               response of HTTP OK (200) is received  from the merchant's server or
#               the number of posts exceeds 10.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet |  object
#========================================================================================
sub PaymentStatusMoneybookers {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop       = $Servlet->object;
    my $FormValues = $Servlet->params();

    $Servlet->vars( 'ViewObjectID', $Shop->id );
    $Servlet->vars( 'ViewAction',   undef );

    $self->_logInformationWithShopContext($Shop, 'PaymentStatusMoneybookers; FormValues', $FormValues);

    #-- Read values posted from Moneybookers payment gateway
#    my $pay_to_email      = $FormValues->{'pay_to_email'};
#    my $pay_from_email    = $FormValues->{'pay_from_email'};
    my $merchant_id       = $FormValues->{'merchant_id'};
    my $transaction_id    = $FormValues->{'transaction_id'};
    my $mb_transaction_id = $FormValues->{'mb_transaction_id'};
    my $mb_amount         = $FormValues->{'mb_amount'};
#    my $mb_currency       = $FormValues->{'mb_currency'};
    my $status            = $FormValues->{'status'};
    my $md5sig            = $FormValues->{'md5sig'};
#    my $amount            = $FormValues->{'amount'};
#    my $currency          = $FormValues->{'currency'};
    my $lineitemcontainer_id  = $FormValues->{'lineitemcontainer_id'};

    # test Container and Shop....
    Error( 'MoneybookersNoBasket' => { 'TransactionID' => $lineitemcontainer_id,
                                       'ShopID' => $Shop->id(),
                                      } ) unless ExistsObject($lineitemcontainer_id);

    my $IsBasketOK = 1;
    #the $lineitemcontainer_id is always the lineitemcontainerid
    #the $transaction_id can be the lineitemcontainerid or the orderid
    my $LineItemContainer = $Shop->loadObject($lineitemcontainer_id, 'LineItemContainer');
    my $Order;
    my $Basket;

    if ( $LineItemContainer->parent()->class()->alias() eq 'CustomerOrder' ) {
        $Order = $LineItemContainer->parent();
    }
    else {
        $Basket = $LineItemContainer->parent();
    }

    $IsBasketOK = 0 unless $LineItemContainer->class()->alias() eq 'LineItemContainer';
    $IsBasketOK = 0 unless $LineItemContainer->getSite() eq $Shop;

    #-- get PaymentLineItem object
    my $LineItemPayment = $LineItemContainer->get('Payment');
    my $PaymentMethod   = $LineItemPayment->get('PaymentMethod');

    #-- Check that $LineItemContainer has/had Moneybookers as payment method
    $IsBasketOK = 0 unless $PaymentMethod->class()->alias() eq 'PaymentMethodMoneybookers';

    Error( 'MoneybookersNoBasket' => { 'TransactionID' => $lineitemcontainer_id,
                                       'ShopID' => $Shop->id(),
                                      } ) unless $IsBasketOK;

    #-- calculate MD5 checksum and compare with $md5sig
    # we compare the MerchantID, LineItemContainerID, CurrencyID. the other params we do not have at this point

    # we can only check the merchant signature for which we hold the secret word
    if ( $merchant_id eq $PaymentMethod->get('MerchantID') ) {
        my $secrect_key = uc(md5_hex( encode_utf8($PaymentMethod->get('Secret') )));
        my $my_md5      = uc(md5_hex( $PaymentMethod->get('MerchantID') . $transaction_id . $secrect_key . $mb_amount
                                    . $LineItemContainer->get('CurrencyID') . $status )
                                    );
        #-- the security check.
        if ( $my_md5 ne $md5sig ) {
            my $hWarningMismatch = {
                'MBTransactionID' => $mb_transaction_id,
                'TransactionID' => $lineitemcontainer_id,
                'Signature' => $md5sig,
                'Mac'       => $my_md5,
                'ShopID'    => $Shop->id(),
            };

            $self->_logInformationWithShopContext($Shop, "Signature Mismatch Warning:\n", $hWarningMismatch);

            Warning( 'MoneybookersSignatureMismatch' => $hWarningMismatch );
        }
    }

    #-- set payment status information to the PaymentLineItem object
    $LineItemPayment->set({'TransID' => $mb_transaction_id});
    $LineItemPayment->set({'TransTime' => GetCurrentDateTime()});

    my $Response = $Servlet->response();
    if( defined $Order ){
        _SetOrderStatus($Servlet,$Order,$FormValues);

        $Response->code( RC_OK );
        $Response->header( 'Location' => $Order->getSite()->get('WebUrl'));
        $Response->content( 'Payment complete: order exists ....' );
        $Servlet->params('ExitAfterEvent', 1);
    }
    else {
        # create the order
        my $OrderGUID = DAL_UnCachedInfoBasket($Basket->id)->{'OrderGUID'};
        if( defined $OrderGUID ) {
            if (!ExistsObjectByGUID($OrderGUID)) {
                # order is beeing created by another process
                # return a 409 (Conflict) to the payment system because the payment call 10 times this function
                # until a 200 is received
                $Response->code( RC_CONFLICT );
                $Response->header( 'Location' => $Basket->getSite()->get('WebUrl'));
                $Response->content( 'Wait while order is created.' );
                $Servlet->params('ExitAfterEvent', 1);
                return;
            }
        }
        # create the order
        if (defined($status) && ($status eq '2' || $status eq '0')) {
            $Order = DE_EPAGES::Order::UI::Basket->doBasket2Order($Servlet, $Basket);
            _SetOrderStatus($Servlet,$Order,$FormValues);
            $Response->code( RC_OK );
            $Response->header( 'Location' => $Order->getSite()->get('WebUrl'));
            $Response->content( 'Payment complete: order created....' );
            $Servlet->params('ExitAfterEvent', 1);
        }
        else {
            $LineItemPayment->set({'TransStatus' => $status});
            $Response->code(RC_OK);
            $Response->header('Location' => $Basket->getSite()->get('WebUrl'));
            $Response->content('Payment declined: no order created....');
            $Servlet->params('ExitAfterEvent', 1);
        }
    }

    return;
}

#========================================================================================
# function     TransactionPaymentName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Name = $Handler->TransactionPaymentName();
#----------------------------------------------------------------------------------------
# description  return Moneybooker as string. Name of PaymentType for search/view
#                               transactions.
#----------------------------------------------------------------------------------------
# return       $Name | name of payment type | string
#========================================================================================
sub TransactionPaymentName { 'Moneybookers' }

#========================================================================================
sub _SetOrderStatus {
    my ($Servlet, $Order, $hParams) = @_;

    my $LineItemContainer = $Order->get('LineItemContainer');
    my $LineItemPayment = $LineItemContainer->get('Payment');

    eval {

        if ($hParams->{'status'} == 2 and $hParams->{'amount'} == $LineItemContainer->get('GrandTotal')) {
            $Order->set({'PaidOn' => GetCurrentDateTime()});
            $LineItemPayment->set({'TransStatus' => 2});
        }
        if ($hParams->{'status'} == 2 and $hParams->{'amount'} != $LineItemContainer->get('GrandTotal') ) {
            $Order->set({'PartlyPaidOn' => GetCurrentDateTime()});
            $LineItemPayment->set({'TransStatus' => 2});
        }
        if ($hParams->{'status'} == 0 or $hParams->{'status'} == 1) {
            $Order->set({'PendingOn' => GetCurrentDateTime()});
            $LineItemPayment->set({'TransStatus' => 0});
        }
        if ($hParams->{'status'} == -1 or $hParams->{'status'} == -2 or $hParams->{'status'} == -3) { ## no critic (ProhibitMagicNumbers)
            $Order->set({'CancelledOn' => GetCurrentDateTime()});
            $LineItemPayment->set({'TransStatus' => $hParams->{'status'}});
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ($Error->code eq 'SEND_MAIL_FAILED') {
            $Servlet->vars('SEND_MAIL_FAILED',1);
        } else {
            Error($Error);
        }
    }
    return;
}

#========================================================================================
sub _validateCallParametersAndGetContainer {
    my $self = shift;
    my ($Shop, $Form, $ServletParams) = @_;

    my $LineItemContainerID = $ServletParams->{'LICID'};

    #-- show error, if the LineItemContainer can't be found
    Error('MoneybookersNoBasket' => { 'TransactionID' => $LineItemContainerID, 'ShopID' => $Shop->id(),
    }) unless $LineItemContainerID;

    Error('MoneybookersNoBasket' => { 'TransactionID' => $LineItemContainerID, 'ShopID' => $Shop->id(),
    }) unless ExistsObject($LineItemContainerID);

    #-- load LineItemContainer object
    my $LineItemContainer = $Shop->loadObject($LineItemContainerID, 'LineItemContainer');

    #-- Check that $LineItemContainer belongs to current shop
    $Form->executeFormError({
        'Name'       => 'PaymentResult',
        'Reason'     => 'BadCallback_ShopIDMisMatch',
        'ViewAction' => 'ViewMoneybookersError',
    }) if (not $LineItemContainer->getSite() eq $Shop);

    return $LineItemContainer;
}

#========================================================================================
sub _logInformationWithShopContext {
    my $self = shift;
    my ( $Shop, $LogDescription, $DataToLog) = @_;

    my $MethodPath = (caller(0))[3];
    my $Tools = DE_EPAGES::Moneybookers::API::Tools->new();
    $Tools->LogShopContext(
            $Shop,
            $MethodPath,
            sub { LogPayment('Moneybookers', $LogDescription, $DataToLog); }
    );
}

1;

