#========================================================================================
# package      DE_EPAGES::Moneybookers::t::unit::UI_Shop
# base         DE_EPAGES::Test::API::UnitTest::Test
# state        public
#----------------------------------------------------------------------------------------
# description  unit test for UI::Shop
#========================================================================================
package DE_EPAGES::Moneybookers::t::unit::UI_Shop;
use base DE_EPAGES::Test::API::UnitTest::Test;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;
use Test::Mockify::Matcher qw ( Object );
use Test::Mockify::Sut;
use Test::Mockify::Verify qw ( WasCalled );
use Test::Exception;
use Test::More;

use strict;

use constant LINE_ITEM_CONTAINER_ID => '34571';
use constant BASKET_ID => '34570';
use constant ORDER_ID => '34587';
use constant USER => 'Anonymous_14';


sub test {
    my $self = shift;

    $self->test_PaymentMoneybookersCallback_on_order_unity_shop();
    $self->test_PaymentMoneybookersCallback_on_basket_ep6_shop();
    $self->test_PaymentMoneybookersCancel_on_order_unity_shop();
    $self->test_PaymentMoneybookersCancel_on_basket_ep6_shop();

    done_testing();
}


sub test_PaymentMoneybookersCallback_on_order_unity_shop {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    ## GIVEN
    my $Order = $self->_buildOrder();
    my $LineItemContainer = $self->_buildLineItemContainerWithParent($Order);
    my $Shop = $self->_buildShop(
        $LineItemContainer,
        { 'IsUnityRegistered' => 1, 'NameOrAlias' => 'unityTestShop' }
    );
    my $Servlet = $self->_buildServlet($Shop);
    my $SUT = $self->_createSUT();

    ## WHEN
    $SUT->PaymentCallbackMoneybookers($Servlet);

    ## THEN
    is(
        $LineItemContainer->get('Payment')->get('PaymentErrorCode'),
        undef,
        "$SubTestName - verify PaymentErrorCode is not set on success."
    );
    is($Servlet->vars('ViewObjectID'), ORDER_ID, "$SubTestName - verify servlet view object id is set with order.");
    is(
        $Servlet->vars('ViewAction'),
        'ViewMultiCheckoutCustomerOrder',
        "$SubTestName - verify servlet view action is set to redirect to unity confirmation."
    );
}

sub test_PaymentMoneybookersCallback_on_basket_ep6_shop {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    ## GIVEN
    my $Basket = $self->_buildBasket();
    my $LineItemContainer = $self->_buildLineItemContainerWithParent($Basket);
    my $Shop = $self->_buildShop(
        $LineItemContainer,
        { 'IsUnityRegistered' => 0, 'NameOrAlias' => 'baseTestShop' }
    );
    my $Servlet = $self->_buildServlet($Shop);
    my $SUT = $self->_createSUT();

    ## WHEN
    $SUT->PaymentCallbackMoneybookers($Servlet);

    ## THEN
    ok(WasCalled(
        $SUT,
        'DE_EPAGES::Order::UI::Basket::doBasket2Order'),
        "$SubTestName - create order when on successful order after payment."
    );
    is(
        $Servlet->vars('ViewObjectID'),
        ORDER_ID,
        "$SubTestName - verify servlet view object id is set with created order."
    );
    is($Servlet->vars('ViewAction'), undef, "$SubTestName - verify servlet view action is set undefined.");

}

sub test_PaymentMoneybookersCancel_on_order_unity_shop {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    ## GIVEN
    my $Order = $self->_buildOrder();
    my $LineItemContainer = $self->_buildLineItemContainerWithParent($Order);
    my $Shop = $self->_buildShop(
        $LineItemContainer,
        { 'IsUnityRegistered' => 1, 'NameOrAlias' => 'unityTestShop' }
    );
    my $Servlet = $self->_buildServlet($Shop);
    my $SUT = $self->_createSUT();

    ## WHEN
    $SUT->PaymentMoneybookersCancel($Servlet);

    ## THEN
    is(
        $LineItemContainer->get('Payment')->get('PaymentErrorCode'),
        'Canceled',
        "$SubTestName - verify PaymentErrorCode is set as canceled."
    );
    is($Servlet->vars('ViewObjectID'), ORDER_ID, "$SubTestName - verify servlet view object id is set with order.");
    is(
        $Servlet->vars('ViewAction'),
        'ViewMultiCheckoutCustomerOrder',
        "$SubTestName - verify servlet view action is set to redirect to unity confirmation."
    );

    return;
}

sub test_PaymentMoneybookersCancel_on_basket_ep6_shop {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    ## GIVEN
    my $Basket = $self->_buildBasket();
    my $LineItemContainer = $self->_buildLineItemContainerWithParent($Basket);
    my $Shop = $self->_buildShop(
        $LineItemContainer,
        { 'IsUnityRegistered' => 0, 'NameOrAlias' => 'baseTestShop' }
    );
    my $Servlet = $self->_buildServlet($Shop);
    my $SUT = $self->_createSUT();

    ## WHEN
    $SUT->PaymentMoneybookersCancel($Servlet);

    ## THEN
    is(
        $LineItemContainer->get('Payment')->get('PaymentErrorCode'),
        undef,
        "$SubTestName - verify PaymentErrorCode is not set for base shops."
    );
    is($Servlet->vars('ViewObjectID'), BASKET_ID, "$SubTestName - verify servlet view object id is set with basket.");
    is(
        $Servlet->vars('ViewAction'),
        'ViewPaymentMoneybookersCancel',
        "$SubTestName - verify servlet view action is set to redirect to epages6 cancel page."
    );

    return;
}

#===============================================================
# Mocks
#===============================================================
sub _createSUT {
    my $self = shift;

    my $aParameterList = [];
    my $MockifySUT = Test::Mockify::Sut->new('DE_EPAGES::Moneybookers::UI::Shop', $aParameterList);
    $MockifySUT->mock('_logInformationWithShopContext')->whenAny()->thenReturnUndef();
    $MockifySUT->mockImported('DE_EPAGES::Core::API::Error', 'Error')->whenAny()->thenReturnUndef();
    $MockifySUT->mockImported('DE_EPAGES::Object::API::Factory', 'ExistsObject')->whenAny()->thenReturn(1);
    $MockifySUT->mockImported('DE_EPAGES::Order::DAL::Basket', 'DAL_UnCachedInfoBasket')
        ->whenAny()->thenReturn( { 'OrderGUID' => undef } );
    $MockifySUT->mockStatic('DE_EPAGES::Order::UI::Basket::doBasket2Order')
        ->when(
            Object('DE_EPAGES::Presentation::Servlet::Permission'),
            Object('DE_EPAGES::Order::API::Object::Basket')
        )->thenReturn( $self->_buildOrder() );

    return $MockifySUT->getMockObject();
}

sub _buildShop {
    my $self = shift;
    my ($LineItemContainer, $hOptions) = @_;

    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Shop::API::Object::Shop',
        { 'IsUnityRegistered' => $hOptions->{'IsUnityRegistered'}, 'NameOrAlias' => $hOptions->{'NameOrAlias'} }
    );
    $Mockify->addMockWithReturnValueAndParameterCheck(
        'loadObject', $LineItemContainer, [ { 'string' => LINE_ITEM_CONTAINER_ID }, { 'string'=>'LineItemContainer' }]
    );

    my $MockedShop = $Mockify->getMockObject();
    $LineItemContainer->set({'Site' => $MockedShop});

    return $MockedShop;
}

sub _buildLineItemContainerWithParent {
    my $self = shift;
    my ($Parent) = @_;

    my $hAttibutes = {
        'Parent' => $Parent,
        'Payment' => $self->_buildLineItemPayment(),
        'Site'    => undef
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Order::API::Object::LineItemContainer',
        $hAttibutes
    );

    return $Mockify->getMockObject();
}

sub _buildOrder {
    my $self = shift;

    my $hAttributes = {
        'ID'         => ORDER_ID,
        'instanceOf' => 'CustomerOrder',
        'User'       => USER
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Order::API::Object::CustomerOrder',
        $hAttributes
    );

    return $Mockify->getMockObject();
}

sub _buildBasket {
    my $self = shift;

    my $hAttributes = {
        'ID' => BASKET_ID,
        'instanceOf' => 'Basket'
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Order::API::Object::Basket',
        $hAttributes
    );

    return $Mockify->getMockObject();
}

sub _buildLineItemPayment {
    my $self = shift;

    my $hAttibutes = {
        'PaymentMethod'    => $self->_buildPaymentMethod(),
        'PaymentErrorCode' => undef,
        'TransStatus'      => 0
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Moneybookers::API::Object::LineItemPaymentMoneybookers',
        $hAttibutes
    );

    return $Mockify->getMockObject();
}

sub _buildPaymentMethod {
    my $self = shift;

    my $hAttributes = {
        'ID' => '12345',
        'instanceOf' => 'PaymentMethodMoneybookers',
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Moneybookers::API::Object::PaymentMethodMoneybookers',
        $hAttributes
    );
    $Mockify->addMockWithReturnValue('pathString', '/payment/method/path');

    return $Mockify->getMockObject();
}

sub _buildServlet {
    my $self = shift;
    my ($Shop) = @_;

    my $MockifyServlet = DE_EPAGES::Test::Mock::MockifyServlet->new($Shop);
    $MockifyServlet->mock('form', $self->_buildForm());
    $MockifyServlet->mock('registerSession', sub { return; });
    $MockifyServlet->addParam('LICID', LINE_ITEM_CONTAINER_ID);

    return $MockifyServlet->getMockObject();
}

sub _buildForm {
    my $self = shift;

    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Presentation::API::Form', []);
    $Mockify->mock('executeFormError', undef, ['hashref']);

    return $Mockify->getMockObject();
}


__PACKAGE__->runScript();