#========================================================================================
# package      DE_EPAGES::Order::API::BasketPatch
# state        public
#----------------------------------------------------------------------------------------
# description  Implements JSON-PATCH (RFC 6902) for the basket object.
#========================================================================================
package DE_EPAGES::Order::API::BasketPatch;
use base DE_EPAGES::WebService::API::JSONPatch;

use DE_EPAGES::Object::API::Factory qw ( LoadObjectByGUID );
use DE_EPAGES::Core::API::Error qw (Error ExistsError GetError);
use DE_EPAGES::Order::API::BasketAddress;
use DE_EPAGES::Order::API::Constants qw ( CART_PATCH_PARTIALLY_COMPLETE CART_PATCH_COMPLETE  );
use DE_EPAGES::Order::API::ProductValidator qw ( ValidateStockLevelSettings ValidateAvailibilityVisibilitySettings );
use feature qw (switch);

use strict;

## no critic (ProhibitUnusedPrivateSubroutines)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( $Basket );
#----------------------------------------------------------------------------------------
# description  Constructor
#----------------------------------------------------------------------------------------
# input        $Basket | A basket | DE_EPAGES::Order::API::Object::Basket
# return       $self | self | DE_EPAGES::Order::API::BasketPatch
#========================================================================================
sub new {
    my $class = shift;
    my ($Basket) = @_;
    my $self  = bless {
        'Basket' => $Basket,
    }, $class;
    return $self;
}

sub _add {
    my $self = shift;
    my ($Path, $hValue, $hParams) = @_;

    if($Path =~ m|^/lineItemContainer/productLineItems/([^/]+)/?$|) {
        my $LineItemGUID = $1;
        return $self->_updateSpecificProductLineItem($hValue, $LineItemGUID);
    }
    if($Path =~ m|^/lineItemContainer/productLineItems/([^/]+)/product-price/?$|) {
        my $LineItemGUID = $1;
        return $self->_updateProductPriceLineItem($hValue, $LineItemGUID);
    }
    if($Path eq '/lineItemContainer/productLineItems') {
        return $self->_addProductLineItem($hValue, $hParams);
    }
    if($Path eq '/billingAddress') {
        return $self->_setAddress($hValue, 'Billing');
    }
    if($Path eq '/shippingAddress') {
        return $self->_setAddress($hValue, 'Shipping');
    }
    if($Path eq '/coupon'){
        return $self->_setCoupon($hValue);
    }

    Error('BadRequest', {'reason' => "Operation add on $Path is not supported!"});
}

sub _replace {
    my $self = shift;
    my ($Path, $hValue) = @_;

    if($Path eq '/locale'){
        return $self->_setLocale($hValue);
    }

    Error('BadRequest', {'reason' => "Operation add on $Path is not supported!"});
}

sub _remove {
    my $self = shift;
    my ($Path, $Value) = @_;

    if($Path =~ m|^/lineItemContainer/productLineItems/([^/]+)/?$|) {
        my $LineItemGUID = $1;
        return $self->_removeSpecificProductLineItem($LineItemGUID);
    }
    if($Path eq '/billingAddress') {
        return $self->_removeAddress('Billing');
    }
    if($Path eq '/shippingAddress') {
        return $self->_removeAddress('Shipping');
    }
    if($Path eq '/coupon'){
        return $self->_removeCoupon($Value);
    }

    Error('BadRequest', {'reason'=>"Operation remove on $Path is not supported!"});
}

#========================================================================================
# function     _addProductLineItem
# state        private
#----------------------------------------------------------------------------------------
# syntax       _addProductLineItem( $rLineItemData, $hParams );
#----------------------------------------------------------------------------------------
# description  Adds the given product to the basket.
#----------------------------------------------------------------------------------------
# input        $rLineItemData | Info about the product(s) to add | ref.hash or ref.array
# input        $hParams | Info about status is returned by using this hash ref  | ref.hash
#========================================================================================
sub _addProductLineItem {
    my $self = shift;
    my ($rLineItemData, $hParams) = @_;

    my $Basket = $self->{'Basket'};
    my $LineItemContainer = $Basket->container();
    my $Shop = $Basket->getSite();

    # use this indicator for backward compatibility with the 
    # /carts/{cartid}/line-items resource
    # process extended error messages if we have an array as input (/carts/{cartid}/multi-line-items resource)
    my $IsArray;
    my $ahErrors = [];

    my $ahLineItemData;
    if (ref $rLineItemData eq 'ARRAY') {
        $ahLineItemData = $rLineItemData;
        $IsArray = 1;
    } else {
        $ahLineItemData = [$rLineItemData];
    }

    foreach my $hLineItemData (@$ahLineItemData){
        my $ProductGUID = $hLineItemData->{'productId'};
        my $Quantity = $hLineItemData->{'quantity'};

        eval {
            my $Product = $self->_loadShopObjectByGUID($Shop, $ProductGUID, 'Product');
            my $ProductLineItem = $self->_getLineItemByProduct($LineItemContainer, $Product);

            if (defined $ProductLineItem) {
                my $ProductLineItemGUID = $ProductLineItem->guid();
                my $PreviousQuantity    = $ProductLineItem->get('Quantity');
                my $NewQuantity         = $PreviousQuantity + $Quantity;
                my $hUpdateRequestBody = {
                    'quantity' => $NewQuantity,
                };

                $self->_updateSpecificProductLineItem($hUpdateRequestBody,$ProductLineItemGUID);

            } else {

                $self->_ValidateStockLevelSettings($Product, $Quantity);
                $self->_ValidateAvailibilityVisibilitySettings($Product);

                eval {
                    $Product->addToContainer($LineItemContainer, {
                        'Quantity' => $Quantity // 1,
                    });
                };
                if (ExistsError()) {
                    my $Error = GetError();
                    if ( $Error->code() eq 'ProductNoPrice' ) {
                        Error('BadRequest', {'Reason' => 'Price is missing for the product'});
                    } else {
                        Error($Error);
                    }
                }
            }
        };
        # handle errors by a different way if we have an array as input
        if (ExistsError()) {
            my $Error = GetError();
            if ($IsArray) {
                # build the correct message for the failed product
                # and add error messages with GUID to array
                my $Message = $Error->vars()->{'reason'} ? $Error->vars()->{'reason'} : $Error->vars()->{'Reason'};
                $Message = $Error->shortMessage unless $Error->code eq 'BadRequest'; 
                push @$ahErrors, { $ProductGUID => $Message };
            } else {
                # normal error handling
                Error($Error);
            }
        }

        $LineItemContainer->recalculate();
    }

    if (scalar @$ahErrors ) {
        # throw error if we have errors on all items
        if (scalar @$ahErrors eq scalar @$ahLineItemData) {
            Error('BadRequest', {'Reason' => 'No products could be added to the basket'});
        } else {
            # add the status to $hParams for adding later to the basket
            $hParams->{'status'} = CART_PATCH_PARTIALLY_COMPLETE;
        }
    } else {
        # add this status only if we have more than one lineitem (multi-line-items call)
        $hParams->{'status'} = CART_PATCH_COMPLETE if $IsArray;
    }
    return;
}

#========================================================================================
# function     _getLineItemByProduct
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getLineItemByProduct( $LineItemContainer, $Product );
#----------------------------------------------------------------------------------------
# description  Returns the ProductLineItem for a given product if existing.
#----------------------------------------------------------------------------------------
# input        $LineItemContainer | LineItemContainer | DE_EPAGES::Order::API::Object::LineItemContainer
# input        $Product | product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _getLineItemByProduct {
    my $self = shift;
    my ($LineItemContainer, $Product) = @_;

    my $LineItems = $LineItemContainer->get('LineItems');
    my @SameProductLineItems =
        grep { $_->instanceOf('LineItemProduct') and
               defined $_->get('Product') and
               $_->get('Product')->id == $Product->id } @$LineItems;

    return scalar( @SameProductLineItems) ? $SameProductLineItems[0] : undef;
}

#========================================================================================
# function     _updateSpecificProductLineItem
# state        private
#----------------------------------------------------------------------------------------
# syntax       _updateSpecificProductLineItem( $hLineItemData, $hLineItemData );
#----------------------------------------------------------------------------------------
# description  Updates the given productLineItem in the basket.
#----------------------------------------------------------------------------------------
# input        $hLineItemData | Info about the product to add | ref_hash
# input        $LineItemGUID | GUID of the specific line item | string
#========================================================================================
sub _updateSpecificProductLineItem {
    my $self = shift;
    my ($hLineItemData, $LineItemGUID) = @_;

    my $Basket = $self->{'Basket'};
    my $Shop = $Basket->getSite();
    my $LineItemProduct = $self->_loadShopObjectByGUID($Shop, $LineItemGUID, 'LineItemProduct');
    $self->_checkIfLineItemBelongsToBasket($LineItemProduct, $Basket);

    my $Product = $LineItemProduct->get('Product');
    my $Quantity = $hLineItemData->{'quantity'};
    $self->_ValidateStockLevelSettings($Product, $Quantity);
    $self->_ValidateAvailibilityVisibilitySettings($Product);

    $LineItemProduct->set({
        'Quantity' => $Quantity,
    });

    my $LineItemContainer = $Basket->container();
    $LineItemContainer->recalculate();

    return;
}

#========================================================================================
# function     _updateProductPriceLineItem
# state        private
#----------------------------------------------------------------------------------------
# syntax       _updateProductPriceLineItem( $hLineItemData, $hLineItemData );
#----------------------------------------------------------------------------------------
# description  Updates the single price of the given productLineItem in the basket.
#----------------------------------------------------------------------------------------
# input        $hLineItemData | Info about the product to add | ref_hash
# input        $LineItemGUID | GUID of the specific line item | string
#========================================================================================
sub _updateProductPriceLineItem {
    my $self = shift;
    my ($hLineItemData, $LineItemGUID) = @_;

    my $Basket = $self->{'Basket'};
    my $Shop = $Basket->getSite();
    my $LineItemProduct = $self->_loadShopObjectByGUID($Shop, $LineItemGUID, 'LineItemProduct');
    $self->_checkIfLineItemBelongsToBasket($LineItemProduct, $Basket);

    my $ProductBasePrice = $hLineItemData->{'singleItemPrice'};
    return unless (defined $ProductBasePrice);

    my $LineItemContainer = $Basket->container();
    $LineItemProduct->set({
        'BasePrice' => $LineItemContainer->formatter->round($ProductBasePrice)
    });

    $LineItemContainer->recalculate({'BasketPatch' => 1});

    return;
}

#========================================================================================
# function     _removeSpecificProductLineItem
# state        private
#----------------------------------------------------------------------------------------
# syntax       _removeSpecificProductLineItem( $LineItemGUID );
#----------------------------------------------------------------------------------------
# description  Removes the given product line item from the basket.
#----------------------------------------------------------------------------------------
# input        $LineItemGUID | GUID of the specific line item | string
#========================================================================================
sub _removeSpecificProductLineItem {
    my $self = shift;
    my ($LineItemGUID) = @_;

    my $Basket = $self->{'Basket'};
    my $Shop = $Basket->getSite();
    my $LineItemProduct = $self->_loadShopObjectByGUID($Shop, $LineItemGUID, 'LineItemProduct');
    $self->_checkIfLineItemBelongsToBasket($LineItemProduct, $Basket);

    my $LineItemContainer = $Basket->container();
    $LineItemProduct->delete();
    $LineItemContainer->recalculate();

    return;
}

sub _removeCoupon {
    my $self = shift;
    my ($Value) = @_;
    my $Basket = $self->{'Basket'};
    my $Container = $Basket->container;
    my $CouponLineItem = $Container->get('Coupon');
    if(defined $CouponLineItem) {
        if($Value eq $CouponLineItem->guid()) {
            $CouponLineItem->delete();
            $Container->recalculate();
        } else {
            Error('BadRequest', {'Reason' => 'cannot remove coupon, wrong GUID'});
        }
    }
    return;
}

sub _setLocale {
    my $self = shift;
    my ($LocaleID) = @_;

    my $Basket = $self->{'Basket'};
    eval {
        $Basket->container()->set({'LocaleID' => $LocaleID});
        $Basket->get('User')->set({'LocaleID' => $LocaleID});
    };
    if (ExistsError()) {
        my $Error = GetError();
        Error('BadRequest', {'reason'=> $Error->shortMessage }, $Error);
    }

    return;
}

#========================================================================================
# function     _setAddress
# state        private
#----------------------------------------------------------------------------------------
# syntax       _setAddress( $hAddressData, $AddressType );
#----------------------------------------------------------------------------------------
# description  Sets the given address in the basket.
#----------------------------------------------------------------------------------------
# input        $hAddressData | address data | hashref
# input        $AddressType | type of the address (Billing or Shipping) | string
#========================================================================================
sub _setAddress {
    my $self = shift;
    my ($hAddressData, $AddressType) = @_;

    my $BasketAddress = $self->_createBasketAddress($self->{'Basket'});
    $BasketAddress->setAddress($hAddressData, $AddressType);

    $self->{'Basket'}->updateTaxArea();
    $self->{'Basket'}->recalculate();

    return;
}

#========================================================================================
# function     _removeAddress
# state        private
#----------------------------------------------------------------------------------------
# syntax       _removeAddress( $AddressType );
#----------------------------------------------------------------------------------------
# description  Removes the given address type from the basket.
#----------------------------------------------------------------------------------------
# input        $AddressType | type of the address (Billing or Shipping) | string
#========================================================================================
sub _removeAddress {
    my $self = shift;
    my ($AddressType) = @_;

    my $BasketAddress = $self->_createBasketAddress($self->{'Basket'});
    $BasketAddress->removeAddress($AddressType);

    $self->{'Basket'}->updateTaxArea();
    $self->{'Basket'}->recalculate();

    return;
}

#========================================================================================
# function     _checkIfLineItemBelongsToBasket
# state        private
#----------------------------------------------------------------------------------------
# syntax       _checkIfLineItemBelongsToBasket( $LineItem, $Basket );
#----------------------------------------------------------------------------------------
# description  Throws an error if the LineItem does not belong to the basket.
#----------------------------------------------------------------------------------------
# input        $LineItem | A LineItem | DE_EPAGES::Order::API::Object::LineItem
# input        $Basket | A basket | DE_EPAGES::Order::API::Object::Basket
#========================================================================================
sub _checkIfLineItemBelongsToBasket {
    my $self = shift;
    my ($LineItem, $Basket) = @_;

    my $LineItemContainerOfLineItem = $LineItem->parent();
    my $BasketOfLineItem = $LineItemContainerOfLineItem->parent();

    my $LineItemGUID = $LineItem->guid();
    my $BasketGUID = $Basket->guid();
    my $BasketOfLineItemGUID = $BasketOfLineItem->guid();

    if($BasketGUID ne $BasketOfLineItemGUID) {
        Error('BadRequest', {'Reason' => "LineItem $LineItemGUID does not belong to the Basket $BasketGUID"});
    }

    return;
}

#========================================================================================
# function     loadShopObjectByGUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $self->loadShopObjectByGUID( $GUID, $ClassAlias )
#----------------------------------------------------------------------------------------
# description  loads object specified by object GUID and verifies that is exists,
#               belongs to the current shop and is an instance of a certain class.
#               Raises an error if any condition is not fulfilled.
#               Returns the object if all tests passed.
#----------------------------------------------------------------------------------------
# input        $Shop | a shop object | DE_EPAGES::Shop::API::Object::Shop
# input        $GUID | object GUID to load | string
# input        $ClassAlias | alias of a class | string
# return       $Object| object | object
#========================================================================================
sub _loadShopObjectByGUID {
    my $self = shift;
    my ($Shop, $GUID, $ClassAlias) = @_;

    my $ObjectID;
    eval {
        $ObjectID = LoadObjectByGUID($GUID)->id();
    };
    if ( ExistsError() ) {
        my $Error = GetError();
        if ( $Error->code() eq 'NoResults_ObjectByGUID' ) {
            Error('ResourceNotFound', {'GUID' => $GUID})
        } else {
            Error($Error);
        }
    }

    return $Shop->getSite()->loadObject($ObjectID, $ClassAlias // 'Object');
}

sub _createBasketAddress {
    my $self = shift;
    my ($Basket) = @_;
    return DE_EPAGES::Order::API::BasketAddress->new($Basket);
}

sub _ValidateStockLevelSettings {
    my $self = shift;
    my ($Product, $LineItemQuantity) = @_;
    eval {
        ValidateStockLevelSettings($Product, $LineItemQuantity);
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'Low StockLevel' ) {
            my $ProductGUID = $Product->guid();
            Error('BadRequest', {'reason'=>"StockLevel of product is lower then requested (total) quantity= $LineItemQuantity, productId= $ProductGUID"});
        } else {
            Error($Error);
        }
    }
    return;
}

sub _ValidateAvailibilityVisibilitySettings {
    my $self = shift;
    my ($Product) = @_;
    eval {
        ValidateAvailibilityVisibilitySettings($Product);
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'Not AvailableOrVisible' ) {
            my $ProductGUID = $Product->guid();
            Error('BadRequest', {'reason'=>"Product is not available on not visible, productId= $ProductGUID"});
        } else {
            Error($Error);
        }
    }
    return;
}


sub _setCoupon{
    my $self = shift;
    my ($CouponCode) = @_;

    my $Basket = $self->{'Basket'};
    my $Shop = $Basket->getSite();
    my $Folder = $Shop->folder('CouponCodes');
    my $Coupon = $Folder->existsChild($CouponCode) ? $Folder->child($CouponCode) : undef ;
    Error('BadRequest', {'Reason' => $CouponCode.' is not a valid coupon!'}) unless defined $Coupon;
    eval {
        $Coupon->addToContainer($Basket->container, {});
    };
    if (ExistsError()) {
    my $Error = GetError();
        Error('BadRequest', {'reason'=> $Error->shortMessage }, $Error);
    }
    my $LineItemContainer = $Basket->container();
    $LineItemContainer->recalculate();

    return undef;

}

1;
