#========================================================================================
# package      DE_EPAGES::Order::API::ProductValidator
# state        public
#----------------------------------------------------------------------------------------
# description  helper class for stock level shop setting, this is use by the REST API
#========================================================================================
package DE_EPAGES::Order::API::ProductValidator;
use base Exporter;

use DE_EPAGES::Shop::API::Constants qw ( STOCK_NO_ADDTOBASKET STOCK_NOT_VISIBLE ); ## no critic (DE_EPAGES::Style::RequireCorrectImports)
use DE_EPAGES::Core::API::Error qw (Error);

use strict;


our @EXPORT_OK = qw(
    ValidateStockLevelSettings
    ValidateAvailibilityVisibilitySettings
);

#========================================================================================
# function     ValidateStockLevelSettings
# state        public
#----------------------------------------------------------------------------------------
# syntax       ValidateStockLevelSettings( $Product, $RequestQuantity );
#----------------------------------------------------------------------------------------
# description  Check if has enough products on stock, according to the shop setting
#               Throws an error in case of low stock level and setting restrition
#----------------------------------------------------------------------------------------
# input        $Product | Product Object | DE_EPAGES::Product::API::Object::Product
# input        $RequestQuantity | Quantity to be request on the lineItem | string
#========================================================================================
sub ValidateStockLevelSettings {
    my ($Product, $RequestQuantity) = @_;

    return unless defined $Product;

    my $Shop = $Product->getSite();
    if( _StockLevelRestriction($Shop)
        && (not _HasEnoughOnStock($Product, $RequestQuantity))) {
            Error('Low StockLevel');
    }
}

#========================================================================================
# function     ValidateAvailibilityVisibilitySettings
# state        public
#----------------------------------------------------------------------------------------
# syntax       ValidateAvailibilityVisibilitySettings( $Product );
#----------------------------------------------------------------------------------------
# description  Check if products is visible and available
#               Throws an error in case of product is not visible or not available
#----------------------------------------------------------------------------------------
# input        $Product | Product Object | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub ValidateAvailibilityVisibilitySettings {
    my ($Product) = @_;

    return unless defined $Product;

    my $hProductInfo = $Product->get(['IsAvailable', 'IsVisible']);
    Error('Not AvailableOrVisible') unless ($hProductInfo->{'IsAvailable'}  and $hProductInfo->{'IsVisible'}); 
    
}

sub _StockLevelRestriction {
    my ($Shop) = @_;
    return ($Shop->get('ProductLowOnStockAction') == STOCK_NO_ADDTOBASKET
            || $Shop->get('ProductLowOnStockAction') == STOCK_NOT_VISIBLE);
}

sub _HasEnoughOnStock {
    my ($Product, $LineItemQuantity) = @_;
    my $ProductStockLevel = $Product->get('StockLevel');
    return 1 unless defined $ProductStockLevel; # AD-5455

    return $LineItemQuantity <= $ProductStockLevel;
}

1;
