#========================================================================================
# package      DE_EPAGES::Product::Hooks::Shop
# state        public
# description  removes taxclass on declared products of shop which uses removed tax class
#========================================================================================
package DE_EPAGES::Product::Hooks::Shop;

use strict;
use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias LoadObject );
use DE_EPAGES::Object::API::XML::AttributeImportHandler qw( ImportTranslation );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw (ImportObjects);
use DE_EPAGES::Product::API::Cartridge;

#========================================================================================
# function     OnBeforeUpdateShop
#----------------------------------------------------------------------------------------
# syntax       OnBeforeUpdateShop( $hParams );
#----------------------------------------------------------------------------------------
# description  changes tax classes of products if tax matrix of shop was changed.
#               (Hook OBJ_BeforeUpdateShop)
#----------------------------------------------------------------------------------------
# input        $hParams | primary key of tax class, key is
#               <ul>
#                   <li>Object - shop - object
#                   <li>Values
#                       <ul>
#                           <li>TaxMatrix - tax matrix - object
#                       </ul>
#               </ul> | ref.hash.*
#========================================================================================
sub OnBeforeUpdateShop {
    my ($hParams) = @_;

    _ChangeTaxAreaOnProducts($hParams) if exists $hParams->{'Values'}{'TaxMatrix'};
    _ChangeTaxAreaOnProducts($hParams,1) if exists $hParams->{'Values'}{'TaxMatrixDigital'};

    return;
}

sub _ChangeTaxAreaOnProducts {
    my ($hParams,$IsDigital) = @_;

    my $Postfix = $IsDigital ? 'Digital' : '';

    my $Shop = $hParams->{'Object'};
    my $OldTaxMatrix = $Shop->get('TaxMatrix'.$Postfix);
    my $NewTaxMatrix = $hParams->{'Values'}->{'TaxMatrix'.$Postfix};
    return if !defined $OldTaxMatrix || $NewTaxMatrix->id == $OldTaxMatrix->id;

    my $ProductClass = LoadClassByAlias('Product');
    my $aOldTaxClasses = $OldTaxMatrix->get('TaxClasses'.$Postfix);
    foreach my $OldTaxClass (@$aOldTaxClasses) {
        my $NewTaxClass = $NewTaxMatrix->existsChild( $OldTaxClass->alias ) ? $NewTaxMatrix->child( $OldTaxClass->alias ) : $NewTaxMatrix->get('DefaultTaxClass'.$Postfix);
        LoadObject( $_ )->set({'TaxClass'.$Postfix => $NewTaxClass })
            foreach @{$ProductClass->findObjects('ByTaxClass'.$Postfix.'IDShopID', [$OldTaxClass->id, $Shop->id])};
    }

}

#========================================================================================
# function     OnInsertShop
#----------------------------------------------------------------------------------------
# syntax       OnInsertShop( $hParams );
#----------------------------------------------------------------------------------------
# description  Creates the folders 'ProductTypes' and 'Products' when a new shop is
#               created.
#               (Hook OBJ_InsertShop)
#----------------------------------------------------------------------------------------
# input        $hParams | primary key of tax class, key is
#               <ul>
#                   <li>Object - shop - object
#               </ul> | ref.hash.*
#========================================================================================
sub OnInsertShop {
    my ($hParams) = @_;
    my $Shop = $hParams->{'Object'};

    my $Cartridge = DE_EPAGES::Product::API::Cartridge->new;
    ImportObjects($Cartridge->xmlDirectory().'/DefaultShop.xml', $Shop->pathString);
    ImportTranslation($Cartridge->xmlDirectory().'/Translation.DefaultShop',$Shop->pathString);

}

#========================================================================================
# function     OnSaveNumberFormats
#----------------------------------------------------------------------------------------
# syntax       OnSaveNumberFormats( $hParams );
# state        public
#----------------------------------------------------------------------------------------
# description  Evaluates the forms 'SaveNumberFormats_Product' and returns its data.
#               Form errors are not thrown directly but passed to the caller.
#----------------------------------------------------------------------------------------
# input        $hParams | form values which are supplied by Form SaveNumberFormats_Product of class Shop.
#               <ul>
#                   <li>Shop - the current Shop - DE_EPAGES::Shop::API::Object::Shop
#                   <li>Form - form object to evaluate - DE_EPAGES::Presentation::API::Form
#                   <li>Data - form values to return to caller - ref.hash
#               </ul> | ref.hash.*
#========================================================================================
sub OnSaveNumberFormats {
    my ($hParams) = @_;

    my $Shop = $hParams->{'Shop'};
    my $Form = $hParams->{'Form'};
    my $hValues = $hParams->{'Values'};


    my $hFormValues = $Form->form($Shop, 'SaveNumberFormats_Product', 1);

    my $NextProductNumber = $hFormValues->{NextProductNo};
    my $Class = $Shop->get('ProductClass');

    if (scalar @{$Class->findObjects('BySiteIDClassIDAlias', [$Shop->id, $Class->id, $NextProductNumber])}) {
        $Form->addFormError({
            'Form'   => $hFormValues->{'FormName'},
            'Reason' => 'OBJECT_EXISTS',
            'Name'   => 'NextProductNo',
        });
    }

    %$hValues = ( %$hValues, %$hFormValues );
    return;
}

#========================================================================================
# function     BeforeSetDefaultMobileSFModeAttribute
#----------------------------------------------------------------------------------------
# syntax       BeforeSetDefaultMobileSFModeAttribute( $hParams );
#----------------------------------------------------------------------------------------
# description  BeforeSetDefaultMobileSFModeAttribute is triggered, set correct
#               MobileSFMode.
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Shop - shop object | object</li>
#                   <li>DefaultSFMode - default 0 | integer</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnBeforeSetDefaultMobileSFModeAttribute {
    my ($hParams) = @_;

    my $Shop = $hParams->{'Shop'};
    if( $Shop->featureMaxValue('Basket') ) {
        $hParams->{'DefaultSFMode'} = 1
    }
}

#========================================================================================
# function     OnBeforeSaveMobileSFMode
#----------------------------------------------------------------------------------------
# syntax       OnBeforeSaveMobileSFMode( $hParams );
#----------------------------------------------------------------------------------------
# description  OnBeforeSaveMobileSFMode is triggered, check if the shop has the basket
#               feature and set DefaultSFMode value.
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Shop - shop object | object</li>
#                   <li>DefaultSFMode - default 0 | integer</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnBeforeSaveMobileSFMode {
    my ($hParams) = @_;

    my $Shop = $hParams->{'Shop'};
    $hParams->{'DefaultSFMode'} = 1 if $Shop->featureMaxValue('Basket');
}

#========================================================================================
# function     OnBeforeCreateUnityShop
#----------------------------------------------------------------------------------------
# syntax       OnBeforeCreateUnityShop( $hParams );
#----------------------------------------------------------------------------------------
# description  OnBeforeCreateUnityShop is triggered, check if the shop has products out
#               of stock or low on stock and activate the related notification
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>Shop - shop object | object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnBeforeCreateUnityShop {
    my ($hParams) = @_;

    my $Shop = $hParams->{'Object'};
    # import notifications
    my $Cartridge = DE_EPAGES::Product::API::Cartridge->new;
    ImportObjects($Cartridge->xmlDirectory().'/ShopNotifications.xml', $Shop->pathString);

    my $aProducts = $Shop->get('Products');
    my $ProductOutOfStockNotification = $Shop->folder('ShopNotifications')->child('ProductOutOfStock');
    my $ProductLowOnStockNotification = $Shop->folder('ShopNotifications')->child('ProductLowOnStock');
    foreach my $Product (@$aProducts) {
        _checkStockLevelNotifications($Product, $ProductOutOfStockNotification, $ProductLowOnStockNotification);
        if ($Product->countSubProducts > 0) {
            foreach my $SubProduct (@{$Product->subProducts}) {
                _checkStockLevelNotifications($SubProduct, $ProductOutOfStockNotification, $ProductLowOnStockNotification);
            }
        }
    }

    # initialize first steps and default product zoom to yes
    my $hItems = $Shop->get('NextSteps') // {};
    $hItems->{'AddProducts'} = 0;
    $Shop->set(
        {
            'NextSteps' => $hItems,
            'UseProductZoom' => 1
        }
    );

    return;
}

sub _checkStockLevelNotifications {
    my ($Product, $ProductOutOfStockNotification, $ProductLowOnStockNotification) = @_;

    my $hProductInfo = $Product->get(['StockLevel', 'StockLevelAlert']);
    my $StockLevel = $hProductInfo->{'StockLevel'};
    my $MinStockLevel = $hProductInfo->{'StockLevelAlert'};
    # product out of stock
    if (defined $StockLevel && $StockLevel <= 0) {
        $ProductOutOfStockNotification->addObject($Product);
    }
    # product low on stock
    if (defined $MinStockLevel && defined $StockLevel && $StockLevel < $MinStockLevel) {
        $ProductLowOnStockNotification->addObject($Product);
    }
}

1;
