#========================================================================================
# package      DE_EPAGES::Saferpay2::UI::Shop
# base         DE_EPAGES::Order::UI::Shop
# state        public
#----------------------------------------------------------------------------------------
# description  UI Controller for shop
#========================================================================================
package DE_EPAGES::Saferpay2::UI::Shop;
use base DE_EPAGES::Order::UI::Shop;
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Saferpay2::API::Constants qw (
    INTERNAL_CARTRIDGE_NAME
    PAYMENT_TYPE_SAFERPAY2
    AUTHORIZED_PAYMENT_STATUS
    CAPTURED_PAYMENT_STATUS
);
use DE_EPAGES::Saferpay2::API::Tools qw ( HandleTransactionResponse );
use DE_EPAGES::Saferpay2::API::SaferpayService;


use strict;

use constant ALREADY_CAPTURED_ERROR_MESSAGE => 'TRANSACTION_ALREADY_CAPTURED';

#========================================================================================
# function     TransactionPaymentName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Name = $Handler->TransactionPaymentName();
#----------------------------------------------------------------------------------------
# description  Name of PaymentType for search/view transactions.
#----------------------------------------------------------------------------------------
# return       $Name | name of payment type | string
#========================================================================================
sub TransactionPaymentName {
    return INTERNAL_CARTRIDGE_NAME;
}


#========================================================================================
# function     PaymentBacklinkSaferpay2
# state        public
#----------------------------------------------------------------------------------------
# syntax       PaymentBacklinkSaferpay2( $Servlet );
#----------------------------------------------------------------------------------------
# description  Redirect URL for the shopper in case of success or error
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub PaymentBacklinkSaferpay2 {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();

    my $hQueryParams = $Servlet->params();
    my $OrderId = $hQueryParams->{'OrderId'};
    my $LineItemPaymentGUID = $hQueryParams->{'LineItemPaymentGUID'};

    my $Order = $Shop->loadObject($OrderId, 'CustomerOrder');
    my $LineItemPayment = $Order->get('LineItemContainer')->get('Payment');

    $self->_LogPayment('PaymentBacklinkSaferpay2', {
        'OrderId' => $OrderId,
        'LineItemPaymentGUID' => $LineItemPaymentGUID,
    });

    if ( $self->_isPaymentGUIDCorrect($LineItemPayment, $LineItemPaymentGUID) ) {
        $Servlet->vars('ViewObjectID', $OrderId);
        if ($Shop->isUnity()) {
            $Servlet->vars('ViewAction', 'ViewMultiCheckoutCustomerOrder');
        } else {
            $Servlet->vars('ViewAction', 'View');
        }
    };

    my $SaferpayService = DE_EPAGES::Saferpay2::API::SaferpayService->new($LineItemPayment);
    my $Token = $LineItemPayment->get('TransID2');

    if ( defined $LineItemPayment->get('TransStatus') ) {
        $self->_LogPayment('PaymentBacklinkSaferpay2', 'Payment already processed');
    } else {

        my $hResponse = $SaferpayService->DoAssertCall($Token, 'PaymentBacklinkSaferpay2');
        $hResponse = HandleTransactionResponse($hResponse);

        # In case of errors with the assert call, this section will always be reached to display error in storefront
        # because no notification through Callback URL is sent on error.
        if ( $hResponse->{'HasErrorsInResponse'} ) {
            $Servlet->vars('SaferpayPaymentFailed', 1);
            $Servlet->vars('SaferpayPaymentFailedReason', $hResponse->{'ErrorName'});

            if ($Shop->isUnity()) {
                $LineItemPayment->set({'PaymentErrorCode' => 'Failed'});
            }
        }

        $self->_updateTransactionStatus($SaferpayService, $LineItemPayment, $hResponse);
    }

    return;
}


#========================================================================================
# function     PaymentCallbackSaferpay2
# state        public
#----------------------------------------------------------------------------------------
# syntax       PaymentCallbackSaferpay2( $Servlet );
#----------------------------------------------------------------------------------------
# description  Server-server callback from payment provider. Webhook for status change to
#               success
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub PaymentCallbackSaferpay2 {
    my $self = shift;
    my ($Servlet) = @_;

    my $hQueryParams = $Servlet->params();
    my $OrderId = $hQueryParams->{'OrderId'};
    my $LineItemPaymentGUID = $hQueryParams->{'LineItemPaymentGUID'};

    my $Shop;
    my $Order;
    eval {
        $Shop = $Servlet->object();
        $Order = $Shop->loadObject($OrderId, 'CustomerOrder');
        my $LineItemPayment = $Order->get('LineItemContainer')->get('Payment');

        $self->_LogPayment('PaymentCallbackSaferpay2', {
            'OrderId' => $OrderId,
            'LineItemPaymentGUID' => $LineItemPaymentGUID,
        });
        if ( !$self->_isPaymentGUIDCorrect($LineItemPayment, $LineItemPaymentGUID) ) {
            Error('LineItemPaymentGUID_Order_MISMATCH');
        }

        my $SaferpayService = DE_EPAGES::Saferpay2::API::SaferpayService->new($LineItemPayment);
        my $Token = $LineItemPayment->get('TransID2');


        if ( defined $LineItemPayment->get('TransStatus') ) {
            $self->_LogPayment('PaymentCallbackSaferpay2', 'Payment already processed');
        } else {
            my $hResponse = $SaferpayService->DoAssertCall($Token, 'PaymentCallbackSaferpay2');
            $hResponse = HandleTransactionResponse($hResponse);

            $self->_updateTransactionStatus($SaferpayService, $LineItemPayment, $hResponse);
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        $self->_LogPayment('PaymentCallbackSaferpay2 failed', {
            'Error'               => $Error->message(),
            'OrderId'             => $OrderId,
            'LineItemPaymentGUID' => $LineItemPaymentGUID
        });
    }

    $Servlet->params('ContentType', 'text/plain');
    $Servlet->params('ExitAfterEvent', 1);
    $Servlet->writeContentOutput('OK');
    return;
}


sub CaptureTransactions {
    my $self = shift;
    my ($Servlet) = @_;

    my $Form = $Servlet->form;
    my @Transactions = $Form->value('ListObjectID');

    my $Shop = $Servlet->object();

    foreach my $TransactionID (@Transactions) {
        my $LineItemPayment = $Shop->loadObject($TransactionID, 'LineItemPayment');
        my $TransStatus = $LineItemPayment->get('TransStatus');
        my $TransNumber = $LineItemPayment->get('TransID');

        $self->_LogPayment('CaptureTransactions', {
            'transID' => $TransactionID,
            'transStatus' => $TransStatus,
            'transNumber' => $TransNumber,
        });
        #Only capture authorized Transactions
        if ($TransStatus eq AUTHORIZED_PAYMENT_STATUS and $TransNumber) {
            $self->_LogPayment('CaptureTransactions', {$TransactionID => AUTHORIZED_PAYMENT_STATUS});
            my $SaferpayService = DE_EPAGES::Saferpay2::API::SaferpayService->new($LineItemPayment);
            my $hResponse;
            eval {
                $hResponse = $SaferpayService->CaptureTransaction($LineItemPayment->get('TransID'));
            };
            if (ExistsError() || defined($hResponse->{'ErrorName'})) {
                $self->_handleCaptureError($hResponse, $LineItemPayment, $Form, $TransNumber);
            }
            if ($hResponse->{'Status'} eq CAPTURED_PAYMENT_STATUS) {
                $LineItemPayment->updateTransactionStatus($TransNumber, $hResponse->{'Status'}, $hResponse->{'Status'});
            }
        } else {
            $self->_handleCaptureStatusesExceptions($TransStatus, $Form, $TransNumber);
        }
    }
    $Form->executeFormError;
    return;
}

sub _instantCaptureTransaction {
    my $self = shift;
    my ($SaferpayService, $LineItemPayment, $TransNumber) = @_;

    my $hResponse;
    eval {
        $hResponse = $SaferpayService->CaptureTransaction($TransNumber);
    };
    if (ExistsError() || defined($hResponse->{'ErrorName'})) {
        my $ErrorMessage = GetError() // $hResponse->{'ErrorName'};
        $self->_LogPayment('Instant Capture Transaction', {'ERROR' => $ErrorMessage});
        if ($ErrorMessage eq ALREADY_CAPTURED_ERROR_MESSAGE) {
            $LineItemPayment->updateTransactionStatus($TransNumber, CAPTURED_PAYMENT_STATUS, CAPTURED_PAYMENT_STATUS);
        }
    }
    if ($hResponse->{'Status'} eq CAPTURED_PAYMENT_STATUS) {
        $LineItemPayment->updateTransactionStatus($TransNumber, $hResponse->{'Status'}, $hResponse->{'Status'});
    }
}

sub _handleCaptureStatusesExceptions {
    my $self = shift;
    my ($TransStatus, $Form, $TransNumber) = @_;

    if ($TransStatus eq CAPTURED_PAYMENT_STATUS) {
        $self->_addFormError($Form, 'ALREADY_CAPTURED', $TransNumber);
        return;
    }
    if ($TransNumber ne 'NONE') {
        $self->_addFormError($Form, 'WRONG_STATUS', $TransNumber);
        return;
    }
    $self->_addFormError($Form, 'INVALID_TRANSACTION_ID', $TransNumber);
    return;
}

sub _handleCaptureError {
    my $self = shift;
    my ($hResponse, $LineItemPayment, $Form, $TransNumber) = @_;

    my $ErrorMessage = GetError() // $hResponse->{'ErrorName'};
    $self->_LogPayment('CaptureTransactions', {'ERROR' => $ErrorMessage});

    if ($hResponse->{'ErrorName'} eq 'TRANSACTION_ALREADY_CAPTURED') {
        $LineItemPayment->updateTransactionStatus($TransNumber, CAPTURED_PAYMENT_STATUS, CAPTURED_PAYMENT_STATUS);
        $self->_addFormError($Form, 'ALREADY_CAPTURED', $TransNumber);
        return;
    }
    $self->_addFormError($Form, 'CAPTURE_FAILED', $TransNumber);
    return;
}


sub _updateTransactionStatus {
    my $self = shift;
    my ($SaferpayService, $LineItemPayment, $hResponse) = @_;

    $self->_updateLiabilityShift($LineItemPayment, $hResponse);
    $LineItemPayment->updateTransactionStatus(
    $hResponse->{'Transaction'}->{'Id'},
    $hResponse->{'Transaction'}->{'Status'},
    $hResponse->{'Transaction'}->{'ExternalStatus'}
    );

    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');
    if ($PaymentMethod->get('IsDirectCapture') && ($hResponse->{'Transaction'}->{'Status'} eq AUTHORIZED_PAYMENT_STATUS)) {
        $self->_instantCaptureTransaction($SaferpayService, $LineItemPayment, $hResponse->{'Transaction'}->{'Id'});
    }
}


sub _updateLiabilityShift {
    my $self = shift;
    my ($LineItemPayment, $hResponse) = @_;

    if (isValid($hResponse->{'Liability'}) && isValid($hResponse->{'Liability'}->{'LiabilityShift'})) {
        $LineItemPayment->updateLiabilityShift($hResponse->{'Liability'}->{'LiabilityShift'});
    }
}

sub _addFormError {
    my $self = shift;
    my ($Form, $Reason, $TransactionId) = @_;

    $Form->addFormError({
        Reason => $Reason,
        TransactionId => $TransactionId,
        LoopName => 'ListObjectIDs',
        Form => 'Transactions',
        Name => 'ListObjectID'
    });
}


sub _LogPayment {
    my $self = shift;
    my ($Name, $DebugData, $Section, $Filter) = @_;

    return LogPayment(PAYMENT_TYPE_SAFERPAY2, $Name, $DebugData, $Section, $Filter);
}

sub _isPaymentGUIDCorrect {
    my $self = shift;
    my ($LineItemPayment, $LineItemPaymentGUID) = @_;

    return ($LineItemPayment->guid() eq $LineItemPaymentGUID);
}

1;