use strict;
use File::Spec::Functions;
use Getopt::Long;
use String::Random;
use Time::HiRes qw(time);
use 5.010; # Needed because we use state vars

use DE_EPAGES::Cartridge::API::Table::Cartridge qw(ExistsCartridgeByPackage);
use DE_EPAGES::Core::API::Error qw(ExistsError GetError Error);
use DE_EPAGES::Core::API::Log qw(LogError LogContext);
use DE_EPAGES::Core::API::PerlTools qw(ElementOf);
use DE_EPAGES::Core::API::Script qw(RunScript);
use DE_EPAGES::Core::API::Warning qw(Warning);
use DE_EPAGES::Database::API::Connection qw(RunOnStore);
use DE_EPAGES::Object::API::Factory qw(LoadRootObject);
use DE_EPAGES::ShopMigration::API::Constants qw(
    SHOPMIGRATION_FINAL_IN_PROGRESS SHOPMIGRATION_FINAL_FAILED
    EXPORT_IMPORT_ORDER
);
use DE_EPAGES::ShopMigration::API::MigrationShop qw(
    RunPreFlightChecksFinal
    ImportShopDataForMigration
    PreMigrationImportTasksFinal
    PostMigrationImportTasksFinal
);
use DE_EPAGES::WebInterface::API::MessageCenter qw(SynchronizeCache);

##no critic qw( RequireCommentForExportedFunctions ProhibitMultiplePackages RequireArgUnpacking ProhibitPackageVars )
# hack log4perl / cspec.E and cspec.G for easier search in ELK
use DE_EPAGES::WebInterface::API::ApplicationServer;
package DE_EPAGES::WebInterface::API::ApplicationServer;
sub currentRequest {return shift;}
sub headers {return shift;}
sub address {return '127.0.0.1';}
sub port {return '0';}
sub header {return @_[0]->{@_[1]};}
our $FakeAppSrv = bless {}, 'DE_EPAGES::WebInterface::API::ApplicationServer';
$DE_EPAGES::WebInterface::API::ApplicationServer::APPLICATION_SERVER = $FakeAppSrv;
# end of hack
## use critic

package main; ##no critic ( ProhibitMultiplePackages )

sub Main {
    local $| = 1;
    my %Options = ();

    my @GetOpt = qw(help passwd=s storename=s shop=s directory=s output! dopre doimport dopost test only=s);
    $Options{output} = 1;
    $Options{dopre} = 0;
    $Options{doimport} = 0;
    $Options{dopost} = 0;
    usage("Unkown parameter specified.\n") unless GetOptions(\%Options, @GetOpt);
    usage() if $Options{help};
    for (qw(storename shop)) {
        usage(sprintf('Missing parameter -%s', $_)) unless $Options{$_};
    }
    $Options{directory} //= catdir($ENV{EPAGES_WEBSECURE}, 'ShopExport', $Options{storename});
    if (defined $Options{only}) {
        $Options{only} = [ split(/\s*,\s*/, $Options{only}) ];
        foreach my $Type (@{$Options{only}}) {
            usage(sprintf('Invalid type %s', $Type)) unless ElementOf($Type, @{EXPORT_IMPORT_ORDER()});
        }
    }

    my $time1 = time;
    my $hCount = {};
    RunOnStore(
        Store      => $Options{storename},
        DBPassword => $Options{passwd},
        Sub        => sub {
            return unless (ExistsCartridgeByPackage('DE_EPAGES::ShopMigration'));
            $Options{Action} = 'Final';
            $Options{WantedStatus} = SHOPMIGRATION_FINAL_IN_PROGRESS;
            $Options{WantedString} = 'FINAL_IN_PROGRESS';
            $hCount->{FinalMigration} = migrateShops(\%Options);
        },
    );
    my $time2 = time;
    printf "need %.3f seconds\n", $time2 - $time1 if $Options{output};
    eval {SynchronizeCache();};
    Warning(GetError()) if ExistsError();
}


sub migrateShops {
    my ($hOptions) = @_;

    my %Count = (Total => 0, Failed => 0, Done => 0);
    my ($StoreName, $Directory, $ShopAlias, $Output) =
        @{$hOptions}{qw(storename directory shop output)};
    my ($DoPre, $DoImport, $DoPost, $Test) =
        @{$hOptions}{qw(dopre doimport dopost test)};
    my $WantedStatus = $hOptions->{WantedStatus};
    my $WantedString = $hOptions->{WantedString};
    my $aShopList = findSingleShop($ShopAlias, $WantedStatus, $WantedString);
    $Count{Total} = scalar @$aShopList;
    return unless ($Count{Total});
    printf "%sRedo Migration steps for %s\n - PreMigrationImportTasks: %s\n - ImportShopData: %s\n - PostMigrationImportTasks: %s\n",
        $Test ? '[TEST] ' : '', $ShopAlias, $DoPre, $DoImport, $DoPost;
    return if $Test;
    printf "Found %s shops in status %s\n", $Count{Total}, $WantedString if $Output;
    foreach my $Shop (@$aShopList) {
        eval {
            my ($SiteID, $RequestID) = ($Shop->get('GUID'), generateRequestID());
            $FakeAppSrv->{'x-epages-requestid'} = $RequestID;
            $FakeAppSrv->{'x-epages-site'} = $SiteID;
            LogContext({ 'Shop' => $Shop->alias, 'RequestID' => $RequestID }, sub {
                RunPreFlightChecksFinal($Shop);
                if ($DoPre) {
                    printf "Running PreMigrationImportTasks()\n";
                    PreMigrationImportTasksFinal($Shop);
                }
                if ($DoImport) {
                    my @ImportOrder;
                    printf "Running ImportShopDataForMigration()\n";
                    if (defined $hOptions->{only}) {
                        @ImportOrder = grep {ElementOf($_, @{$hOptions->{only}})} @{EXPORT_IMPORT_ORDER()};
                        printf "Limited to: %s\n", join(',', @ImportOrder);
                    }
                    ImportShopDataForMigration($Shop, $Directory, \@ImportOrder);
                }
                if ($DoPost) {
                    printf "Running PostMigrationImportTasks()\n";
                    PostMigrationImportTasksFinal($Shop);
                }
            });
        };
        # Set error state on BaseShop in case something goes wrong so we don't try over and over
        if (ExistsError()) {
            my $Error = GetError();
            $Shop->set({ 'MigrationStatus' => SHOPMIGRATION_FINAL_FAILED, 'IsClosed' => 0 });
            my $Msg = sprintf "%s Migration for shop %s store %s failed with '%s'\n",
                $hOptions->{Action}, $Shop->alias, $StoreName, $Error->shortMessage;
            $hOptions->{errmailcontent} .= $Msg . "\n";
            print $Msg;
            LogError($Error);
            $Count{Failed}++;
        } else {
            printf "%s Migration for shop %s store %s succeeded\n",
                $hOptions->{Action}, $Shop->alias, $StoreName;
            $Count{Done}++;
        }
    }
    return \%Count;
}

sub initializeGenerator {
    my $Generator = String::Random->new;
    $Generator->{'A'} = [ @{$Generator->{'C'}}, @{$Generator->{'c'}}, @{$Generator->{'n'}} ];
    return $Generator;
}

# Generates a RequestID of 27 chars which looks like the Apache unique_id
sub generateRequestID {
    state $Generator = initializeGenerator();
    return $Generator->randpattern('AAAAAAAAAAAAAAAAAAAAAAAAAAA');
}

sub findSingleShop {
    my ($ShopAlias, $WantedStatus, $WantedString) = @_;

    my $ShopFolder = LoadRootObject()->child('Shops');
    usage(sprintf('No such shop: %s', $ShopAlias)) unless $ShopFolder->existsChild($ShopAlias);
    if (!$ShopFolder->existsChild($ShopAlias)) {
        printf("Shop is not existing: %s\n", $ShopAlias);
        exit 2;
    }
    my $Shop = $ShopFolder->child($ShopAlias);
    my ($IsStatus, $IsStatusString) = @{$Shop->get([ 'MigrationStatus', 'MigrationStatusString' ])}
        {qw(MigrationStatus MigrationStatusString)};
    if ($IsStatus != $WantedStatus) {
        printf("Shop %s is not in the proper status (Wanted: %s [%s], Is: %s [%s])\n",
            $ShopAlias, $WantedStatus, $WantedString, $IsStatus, $IsStatusString);
        return [];
    } elsif (!$Shop->get('IsMigrationShop')) {
        printf("Shop is not marked for migration (IsMigrationShop=0)\n", $ShopAlias);
        exit 2;
    }
    return [ $Shop ];
}

sub usage {
    my ($ErrorNote) = @_;

    print STDERR "\n$ErrorNote\n\n" if $ErrorNote;
    print <<END_USAGE;
Repeats steps from the final migration in case some of them failed.
Only works for a single shop.
Please set the MigrationStatus to FINAL_IN_PROGRESS before running
this script.

Usage:
    perl $0 [options] flags objects

options:
    -passwd      database user password                                    ()
    -shop        only use this shop alias instead of all matching shops    required
    -storename   name of store                                             required
    -only        limit "doimport" to this object type, possible values:
                 ShopData MailTypes LegalPages Categories ProductTypes
                 Products CategoryProductMaps CrossSellings                (all)


flags:
    -help        show the command line options                             ()
    -test        run in test mode                                          (0)
    -dopre       run PreMigrationImportTasks (Reset + TransferEpagesUI)    (0)
    -doimport    run ImportShopDataForMigration                            (0)
    -dopost      run PostMigrationImportTasks                              (0)
    -(no)output  turn on more verbose status output                        (1)

Example:
    perl $0 -storename Store -shop DemoShop -doimport -only ShopData,CrossSellings,Products
    perl $0 -storename Store -shop DemoShop -dopre -doimport -dopost
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main,
);

1;
