#========================================================================================
# §package      DE_EPAGES::Spreedly::API::RestClient
# §base         DE_EPAGES::EPagesJ::API::Object::RestClient
# §state        public
# §description  Client for Spreedly
#========================================================================================
package DE_EPAGES::Spreedly::API::RestClient;
use base DE_EPAGES::EPagesJ::API::Object::RestClient;

use strict;

use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Database::API::DataEncryption qw( Encrypt );
use DE_EPAGES::EPagesJ::API::Object::UriTemplate;
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Spreedly::API::Tools;

use DE_EPAGES::Spreedly::API::Constants qw (
    PURCHASE_REQUEST_THREE_DS_VERSION
    PURCHASE_REQUEST_ATTEMPT_3DSECURE
    THREE_DS_AVAILABLE_GATEWAYS
    SPECIFIC_FIELDS_GATEWAYS
);

use constant PAYMENT_TYPE_SPREEDLY_PAYU_LATAM => 'SpreedlyPayULatam';
use constant PAYMENT_TYPE_SPREEDLY_STRIPE_PAYMENT_INTENTS => 'SpreedlyStripePaymentIntents';
use constant PAYMENT_TYPE_SPREEDLY_PAYMENTEZ => 'SpreedlyPaymentez';


use Readonly;
Readonly my %URI_TEMPLATES => (
    PutGateway  => 'payments/spreedly/gateways',
    GetGateway  => 'payments/spreedly/gateways?shopId={shopId}&gatewayName={gatewayName}',
    Purchase    => 'payments/spreedly/purchase?shopId={shopId}&gatewayName={gatewayName}',
    Complete    => 'payments/spreedly/complete?shopId={shopId}&gatewayName={gatewayName}&transactionToken={transactionToken}',
    Show        => 'payments/spreedly/show?shopId={shopId}&gatewayName={gatewayName}&transactionToken={transactionToken}',
);

#========================================================================================
# §function     new
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       DE_EPAGES::Spreedly::API::PaymentRequest->new( %args );
#----------------------------------------------------------------------------------------
# §description  Instanciates a new OAuthTokenClient.
#----------------------------------------------------------------------------------------
# §input        %args | Input arguments. Possible keys.<ul>
#                 <li>AppsUriTemplate - </li>
#                 <li>AppUriTemplate - ... - DE_EPAGES::EPagesJ::API::Object::UriTemplate</li>
#                 <li>LogCategory - category for commlogger. Defaults to "AppStore" - string</li>
#               </ul> | hash
# §return       OAuthTokenClient | new instance | object
#========================================================================================
sub new {
    my $class = shift;
    my %args  = @_;
    $args{LogCategory} //= 'Spreedly';

    return $class->SUPER::new(%args);
}

#========================================================================================
# §function     create
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $TokenClient = DE_EPAGES::Spreedly::API::PaymentRequest->create;
#----------------------------------------------------------------------------------------
# §description  Factory method for a OAuthTokenClient object.
#               Initializes based on the current configuration.
#----------------------------------------------------------------------------------------
# §input        $hArgs | Input arguments. Possible keys.<ul>
#                 <li>BaseURL</li>
#                 <li>ShopType</li>
#                 <li>BusinessUnit</li>
#                 <li>Shop</li> | ref.hash
# §return       $TokenClient | OAuthTokenClient object | DE_EPAGES::OAuthTokenClient::API::OAuth::OAuthTokenClient
#========================================================================================
sub create {
    my $class = shift;
    my ($hArgs) = @_;

    my $System = _LoadRootObject();
    $hArgs->{BaseURL} //= $System->get('OAuthServiceBaseUrl') // $System->get('EPagesJBaseURL');    # TODO: Check which field should we send here.
    $hArgs->{ShopType}     //= $hArgs->{Shop}->get('ShopTypeRef')->alias;
    $hArgs->{Shop}         //= $hArgs->{Shop};
    $hArgs->{BusinessUnit} //= $System->get('ProviderID');
    foreach my $key ( keys %URI_TEMPLATES ) {
        $hArgs->{UriTemplates}->{$key} = DE_EPAGES::EPagesJ::API::Object::UriTemplate->new( $hArgs->{BaseURL}, $URI_TEMPLATES{$key} );
    }

    return $class->SUPER::create( $System, $hArgs );

}

#========================================================================================
# §function     createResource
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $paymentRequest->createResource($gatewayid);
#----------------------------------------------------------------------------------------
# §description  It creates a new gateway resource if it doesn't exist.
#----------------------------------------------------------------------------------------
# §input        $gatewayid | gateway identificator | String
# §input        $Form | form object | object
# §return       $Response | epagesj endpoint url | object
#========================================================================================
sub createResource {
    my $self = shift;
    my ( $GatewayName, $hAdditionalBodyParams ) = @_;

    my $URL             = $self->_createPutGatewayEndpointURL();
    my $hRequestBody    = $self->_buildGatewayRequestBody($GatewayName, $hAdditionalBodyParams);



    my $hEncodedRequest = $self->encodeRequest($hRequestBody);

    my $Response = $self->client->put( $URL, undef, $hEncodedRequest );
    my $hGatewayResponse = $self->_decodeFromResponse($Response);

    return $hGatewayResponse;
}

#========================================================================================
# §function     getResource
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $paymentRequest->getResource($GatewayName);
#----------------------------------------------------------------------------------------
# §description  It gets a gateway resource if it exists.
#----------------------------------------------------------------------------------------
# §input        $GatewayName | gateway identificator | String
# §return       $Response | epagesj endpoint url | object
#========================================================================================
sub getResource {
    my $self = shift;
    my ($GatewayName) = @_;

    my $QueryURL = $self->_createGetGatewayEndpointURL($GatewayName);
    my $Response = $self->client->get($QueryURL, {});

    return $self->decodeResponse($Response->content());
}

sub makePurchase {
    my $self = shift;
    my ( $LineItemContainer, $GatewayName, $PurchaseParams) = @_;

    my $hPurchase = $self->_buildPurchaseData( $LineItemContainer);

    if ($GatewayName ~~ THREE_DS_AVAILABLE_GATEWAYS) {
        $hPurchase = $self->_add3DSToPurchaseData($hPurchase, $PurchaseParams);
    }

    if($GatewayName ~~ SPECIFIC_FIELDS_GATEWAYS) {
        $hPurchase = $self->_addGatewaySpecificFields($hPurchase, $LineItemContainer, $PurchaseParams);
    }

    my $URL = $self->_createPurchaseEndpointURL($GatewayName);

    my $hEncodedPurchase = $self->encodeRequest($hPurchase);
    my $Response = $self->client->post( $URL, undef, $hEncodedPurchase );

    my $hTransaction = $self->_decodeFromResponse($Response);

    return $hTransaction;
}

sub CompleteTransaction {
    my $self = shift;
    my ($GatewayName, $TransactionToken) = @_;

    my $URL = $self->_createCompleteEndpointURL($GatewayName, $TransactionToken);
    my $Response = $self->client->post( $URL, undef, undef );

    my $hTransaction = $self->_decodeFromResponse($Response);

    return $hTransaction;
}

sub ShowTransaction {
    my $self = shift;
    my ($GatewayName, $TransactionToken) = @_;

    my $URL = $self->_createShowEndpointURL($GatewayName, $TransactionToken);
    my $Response = $self->client->get( $URL, {} );

    my $hTransaction = $self->_decodeFromResponse($Response);

    return $hTransaction;
}

#----------------------------------------------------------------------------------------
# Private methods
#----------------------------------------------------------------------------------------
sub _buildGatewayRequestBody {
    my $self = shift;
    my ($GatewayName, $hAdditionalBodyParams) = @_;

    my $hRequest = {
        'shopGUID' => $self->_shop->guid(),
        'name'     => $GatewayName,
    };

     @$hRequest{keys %$hAdditionalBodyParams} = values %$hAdditionalBodyParams;

    return $hRequest;
}

sub _add3DSToPurchaseData {
    my $self = shift;
    my ( $hPurchase, $PurchaseParams ) = @_;

    $hPurchase->{'transaction'}->{'browser_info'} = $PurchaseParams->{'SpreedlyBrowserInfo'};
    $hPurchase->{'transaction'}->{'callback_url'} = $PurchaseParams->{'CallbackUrl'};
    $hPurchase->{'transaction'}->{'redirect_url'} = $PurchaseParams->{'RedirectUrl'};
    $hPurchase->{'transaction'}->{'three_ds_version'} = PURCHASE_REQUEST_THREE_DS_VERSION;
    $hPurchase->{'transaction'}->{'attempt_3dsecure'} = PURCHASE_REQUEST_ATTEMPT_3DSECURE;

    # When developing on Intranet, uncomment following line (because of DNS check):
    # $hPurchase->{'transaction'}->{'callback_url'} = 'https://example.com/callback';

    return $hPurchase;
}

sub _addGatewaySpecificFields {
    my $self = shift;
    my ( $hPurchase, $LineItemContainer, $hPurchaseParams ) = @_;

    my $BasketOrOrder = $LineItemContainer->get('BasketOrOrder');
    my $LineItemPayment = $LineItemContainer->get('Payment');

    return $hPurchase unless defined($LineItemPayment);

    my $LineItemPaymentType = $LineItemPayment->get('PaymentType');
    my $hGatewaySpecificFields = undef;

    if ($LineItemPaymentType->alias() eq PAYMENT_TYPE_SPREEDLY_PAYU_LATAM ) {
        my $BillingAddressBirthDay = $BasketOrOrder->get('BillingAddress')->get('Birthday');
        my $FormattedDate;
        if( isValid( $BillingAddressBirthDay ) ) {
            $FormattedDate = $self->_getBirthdayFormatted($BillingAddressBirthDay);
        } elsif( isValid( $hPurchaseParams->{'GatewayFieldBirthday'})) {
            my $LocaleID = $LineItemContainer->get('LocaleID');
            my $DateTimeBirthday = $self->_parseDateStr( $hPurchaseParams->{'GatewayFieldBirthday'}, $LocaleID );
            $FormattedDate = $self->_getBirthdayFormatted($DateTimeBirthday);
        }
        if(isValid($FormattedDate)) {
            $hGatewaySpecificFields = {
                'payu_latam'  => {
                    'birth_date' => $FormattedDate,
                }
            };

            if( isValid( $hPurchaseParams->{'GatewayFieldDeviceSessionId'})) {
                $hGatewaySpecificFields->{'payu_latam'}->{'device_session_id'} = $hPurchaseParams->{'GatewayFieldDeviceSessionId'};
            }
        }
    } elsif ($LineItemPaymentType->alias() eq PAYMENT_TYPE_SPREEDLY_STRIPE_PAYMENT_INTENTS) {
        my $PaymentMethod   = $LineItemPayment->get('PaymentMethod');
        my $PaymentInformationToken = $LineItemPayment->get('PaymentInformationToken');
        $hGatewaySpecificFields = {
            'stripe_payment_intents'  => {
                'stripe_account' => $PaymentMethod->get('StripeAccountId'),
                'idempotency_key' => Encrypt($PaymentInformationToken)
            }
        };
    } elsif ($LineItemPaymentType->alias() eq PAYMENT_TYPE_SPREEDLY_PAYMENTEZ) {
        my $User = $BasketOrOrder->get('User');
        $hGatewaySpecificFields = {
            'paymentez' => {
                'vat'           => $LineItemContainer->get('TotalTax'), # tax amount with two decimals.
                'user_id'       => $User->alias(), # customer identifier inside your app.
                'dev_reference' => $LineItemContainer->guid() # id from commerce to identify order.
            }
        }
    }

    if(defined $hGatewaySpecificFields) {
        $hPurchase->{'transaction'}->{'gateway_specific_fields'} = $hGatewaySpecificFields;
    }

    return $hPurchase;
}

sub _buildPurchaseData {
    my $self = shift;
    my ( $LineItemContainer) = @_;

    my $LineItemPayment = $LineItemContainer->get('Payment');

    my $PaymentInfoToken = $LineItemPayment->get('PaymentInformationToken');
    my $hOrderAmount = $self->_GetOrderAmount($LineItemContainer);

    my $Tools = DE_EPAGES::Spreedly::API::Tools->new();
    my $hPurchase = {
        'transaction' => {
            'payment_method_token' => $PaymentInfoToken,
            'amount'               => $Tools->convertNumberToSpreedlyFormat( $hOrderAmount->{'amount'} ),
            'currency_code'        => $hOrderAmount->{'currency_code'},
        },
    };

    return $hPurchase;
}

#----------------------------------------------------------------------------------------
# Private methods
#----------------------------------------------------------------------------------------

sub _GetOrderAmount {
    my $self = shift;
    my ($LineItemContainer) = @_;

    my $hLineItemAmount = $LineItemContainer->get(['GrandTotal', 'CurrencyID']);

    my $hOrderAmount = {
        'amount' => $hLineItemAmount->{'GrandTotal'},
        'currency_code' => $hLineItemAmount->{'CurrencyID'},
    };

    return $hOrderAmount;
}

sub _getBirthdayFormatted {
    my $self = shift;
    my ($BirthDay) = @_;

    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new();
    return undef unless ( isValid($BirthDay));
    my $FormattedDate = $Formatter->format_datetime($BirthDay, 'date', '%Y-%m-%d');

    return $FormattedDate;
}

sub _parseDateStr {
    my $self = shift;
    my ($DateStr, $Locale) = @_;

    my $DateTime;
    eval {
        my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new('locale'=> $Locale);
        $DateTime = $Formatter->parse_datetime( $DateStr, 'date' );
    };
    if (ExistsError()) {
        my $Error = GetError();
        GetLog()->warn('Could not convert String to dateTime: '.$Error->message);
    }


    return $DateTime;
}


sub _createPutGatewayEndpointURL {
    my $self = shift;

    return $self->{UriTemplates}->{PutGateway}->uri( {} );
}

sub _createGetGatewayEndpointURL {
    my $self = shift;
    my ($GatewayName) = @_;

    return $self->{UriTemplates}->{GetGateway}->uri( { shopId => $self->_shop->guid(), gatewayName => $GatewayName } );
}

sub _createPurchaseEndpointURL {
    my $self = shift;
    my ($GatewayName) = @_;

    return $self->{UriTemplates}->{Purchase}->uri( { shopId => $self->_shop->guid(), gatewayName => $GatewayName } );
}

sub _createCompleteEndpointURL {
    my $self = shift;
    my ($GatewayName, $TransactionToken) = @_;

    return $self->{UriTemplates}->{Complete}->uri( {
            shopId           => $self->_shop->guid(),
            gatewayName      => $GatewayName,
            transactionToken => $TransactionToken
        } );
}

sub _createShowEndpointURL {
    my $self = shift;
    my ($GatewayName, $TransactionToken) = @_;

    return $self->{UriTemplates}->{Show}->uri( {
        shopId           => $self->_shop->guid(),
        gatewayName      => $GatewayName,
        transactionToken => $TransactionToken
    } );
}

sub _shop {
    my $self = shift;
    return $self->{Shop};
}


#----------------------------------------------------------------------------------------
# Wrappers
#----------------------------------------------------------------------------------------
sub _decodeFromResponse {
    my $self = shift;
    my ($Response) = @_;

    return $self->decodeResponse( $Response->decoded_content() );
}

#----------------------------------------------------------------------------------------

sub _LoadRootObject {
    shift;

    return LoadRootObject();
}

1;

